import itertools
import logging
from copy import copy
from typing import Set, List, KeysView, Union, Dict, Optional, Tuple, Any

log = logging.getLogger('assemblyline.classification')


class InvalidClassification(Exception):
    pass


class InvalidDefinition(Exception):
    pass


class Classification(object):
    MIN_LVL = 1
    MAX_LVL = 10000
    NULL_LVL = 0
    INVALID_LVL = 10001
    NULL_CLASSIFICATION = "NULL"
    INVALID_CLASSIFICATION = "INVALID"

    def __init__(self, classification_definition: Dict):
        """
        Returns the classification class instantiated with the classification_definition

        Args:
            classification_definition:  The classification definition dictionary,
                                        see default classification.yml for an example.
        """
        banned_params_keys = ['name', 'short_name', 'lvl', 'aliases', 'auto_select', 'css', 'description']
        self.original_definition = classification_definition
        self.levels_map = {}
        self.levels_map_stl = {}
        self.levels_map_lts = {}
        self.levels_styles_map = {}
        self.levels_aliases = {}
        self.access_req_map_lts = {}
        self.access_req_map_stl = {}
        self.access_req_aliases = {}
        self.groups_map_lts = {}
        self.groups_map_stl = {}
        self.groups_aliases = {}
        self.groups_auto_select = []
        self.groups_auto_select_short = []
        self.subgroups_map_lts = {}
        self.subgroups_map_stl = {}
        self.subgroups_aliases = {}
        self.subgroups_auto_select = []
        self.subgroups_auto_select_short = []
        self.params_map = {}
        self.description = {}
        self.invalid_mode = False
        self._classification_cache = set()
        self._classification_cache_short = set()

        self.enforce = False
        self.dynamic_groups = False

        # Add Invalid classification
        self.levels_map["INV"] = self.INVALID_LVL
        self.levels_map[str(self.INVALID_LVL)] = "INV"
        self.levels_map_stl["INV"] = self.INVALID_CLASSIFICATION
        self.levels_map_lts[self.INVALID_CLASSIFICATION] = "INV"

        # Add null classification
        self.levels_map[self.NULL_CLASSIFICATION] = self.NULL_LVL
        self.levels_map[str(self.NULL_LVL)] = self.NULL_CLASSIFICATION
        self.levels_map_stl[self.NULL_CLASSIFICATION] = self.NULL_CLASSIFICATION
        self.levels_map_lts[self.NULL_CLASSIFICATION] = self.NULL_CLASSIFICATION

        try:
            self.enforce = classification_definition['enforce']
            self.dynamic_groups = classification_definition['dynamic_groups']
            if self.enforce:
                self._classification_cache = self.list_all_classification_combinations()
                self._classification_cache_short = self.list_all_classification_combinations(long_format=False)

            for x in classification_definition['levels']:
                short_name = x['short_name'].upper()
                name = x['name'].upper()

                if short_name in ["INV", "NULL"] or name in [self.INVALID_CLASSIFICATION, self.NULL_CLASSIFICATION]:
                    raise InvalidDefinition("You cannot use reserved words NULL, INVALID or INV in your "
                                            "classification definition.")

                lvl = int(x['lvl'])
                if lvl > self.MAX_LVL:
                    raise InvalidDefinition("Level over maximum classification level of %s." % self.MAX_LVL)
                if lvl < self.MIN_LVL:
                    raise InvalidDefinition("Level under minimum classification level of %s." % self.MIN_LVL)

                self.levels_map[short_name] = lvl
                self.levels_map[str(lvl)] = short_name
                self.levels_map_stl[short_name] = name
                self.levels_map_lts[name] = short_name
                for a in x.get('aliases', []):
                    self.levels_aliases[a.upper()] = short_name
                self.params_map[short_name] = {k: v for k, v in x.items() if k not in banned_params_keys}
                self.params_map[name] = self.params_map[short_name]
                self.levels_styles_map[short_name] = x.get('css', {'banner': 'alert-default',
                                                                   'label': 'label-default',
                                                                   'text': 'text-muted'})
                self.levels_styles_map[name] = self.levels_styles_map[short_name]
                self.description[short_name] = x.get('description', "N/A")
                self.description[name] = self.description[short_name]

            for x in classification_definition['required']:
                short_name = x['short_name'].upper()
                name = x['name'].upper()
                self.access_req_map_lts[name] = short_name
                self.access_req_map_stl[short_name] = name
                for a in x.get('aliases', []):
                    self.access_req_aliases[a.upper()] = self.access_req_aliases.get(a.upper(), []) + [short_name]
                self.params_map[short_name] = {k: v for k, v in x.items() if k not in banned_params_keys}
                self.params_map[name] = self.params_map[short_name]
                self.description[short_name] = x.get('description', "N/A")
                self.description[name] = self.description[short_name]

            for x in classification_definition['groups']:
                short_name = x['short_name'].upper()
                name = x['name'].upper()
                self.groups_map_lts[name] = short_name
                self.groups_map_stl[short_name] = name
                for a in x.get('aliases', []):
                    self.groups_aliases[a.upper()] = \
                        list(set(self.groups_aliases.get(a.upper(), []) + [short_name]))
                solitary_display_name = x.get('solitary_display_name', None)
                if solitary_display_name:
                    self.groups_aliases[solitary_display_name.upper()] = \
                        list(set(self.groups_aliases.get(solitary_display_name.upper(), []) + [short_name]))
                if x.get('auto_select', False):
                    self.groups_auto_select.append(name)
                    self.groups_auto_select_short.append(short_name)
                self.params_map[short_name] = {k: v for k, v in x.items() if k not in banned_params_keys}
                self.params_map[name] = self.params_map[short_name]
                self.description[short_name] = x.get('description', "N/A")
                self.description[name] = self.description[short_name]

            for x in classification_definition['subgroups']:
                short_name = x['short_name'].upper()
                name = x['name'].upper()
                self.subgroups_map_lts[name] = short_name
                self.subgroups_map_stl[short_name] = name
                for a in x.get('aliases', []):
                    self.subgroups_aliases[a.upper()] = \
                        list(set(self.subgroups_aliases.get(a.upper(), []) + [short_name]))
                solitary_display_name = x.get('solitary_display_name', None)
                if solitary_display_name:
                    self.subgroups_aliases[solitary_display_name.upper()] = \
                        list(set(self.subgroups_aliases.get(solitary_display_name.upper(), []) + [short_name]))
                if x.get('auto_select', False):
                    self.subgroups_auto_select.append(name)
                    self.subgroups_auto_select_short.append(short_name)
                self.params_map[short_name] = {k: v for k, v in x.items() if k not in banned_params_keys}
                self.params_map[name] = self.params_map[short_name]
                self.description[short_name] = x.get('description', "N/A")
                self.description[name] = self.description[short_name]

            if not self.is_valid(classification_definition['unrestricted']):
                raise InvalidDefinition("Classification definition's unrestricted classification is invalid.")

            if not self.is_valid(classification_definition['restricted']):
                raise InvalidDefinition("Classification definition's restricted classification is invalid.")

            self.UNRESTRICTED = classification_definition['unrestricted']
            self.RESTRICTED = classification_definition['restricted']

            self.UNRESTRICTED = self.normalize_classification(classification_definition['unrestricted'])
            self.RESTRICTED = self.normalize_classification(classification_definition['restricted'])

        except Exception as e:
            self.UNRESTRICTED = self.NULL_CLASSIFICATION
            self.RESTRICTED = self.INVALID_CLASSIFICATION

            self.invalid_mode = True

            log.warning(str(e))

    ############################
    # Private functions
    ############################
    @staticmethod
    def _build_combinations(items: Set, separator: str = "/", solitary_display: Optional[Dict] = None) -> Set:
        if solitary_display is None:
            solitary_display = {}

        out = {""}
        for i in items:
            others = [x for x in items if x != i]
            for x in range(len(others)+1):
                for c in itertools.combinations(others, x):
                    value = separator.join(sorted([i]+list(c)))
                    out.add(solitary_display.get(value, value))

        return out

    @staticmethod
    def _list_items_and_aliases(data: List, long_format: bool = True) -> Set:
        items = set()
        for item in data:
            if long_format:
                items.add(item['name'])
            else:
                items.add(item['short_name'])

        return items

    def _get_c12n_level_index(self, c12n: str) -> str:
        # Parse classifications in uppercase mode only
        c12n = c12n.upper()

        lvl = c12n.split("//")[0]
        if lvl in self.levels_map:
            return self.levels_map[lvl]
        elif lvl in self.levels_map_lts:
            return self.levels_map[self.levels_map_lts[lvl]]
        elif lvl in self.levels_aliases:
            return self.levels_map[self.levels_aliases[lvl]]
        else:
            raise InvalidClassification("Classification level '%s' was not found in "
                                        "your classification definition." % lvl)

    def _get_c12n_level_text(self, lvl_idx: int, long_format: bool = True) -> str:
        text = self.levels_map.get(str(lvl_idx), None)
        if not text:
            raise InvalidClassification("Classification level number '%s' was not "
                                        "found in your classification definition." % lvl_idx)
        if long_format:
            return self.levels_map_stl[text]
        return text

    def _get_c12n_required(self, c12n: str, long_format: bool = True) -> List:
        # Parse classifications in uppercase mode only
        c12n = c12n.upper()

        return_set = set()
        part_set = set(c12n.split("/"))

        for p in part_set:
            if p in self.access_req_map_lts:
                return_set.add(self.access_req_map_lts[p])
            elif p in self.access_req_map_stl:
                return_set.add(p)
            elif p in self.access_req_aliases:
                for a in self.access_req_aliases[p]:
                    return_set.add(a)

        if long_format:
            return sorted([self.access_req_map_stl[r] for r in return_set])
        return sorted(list(return_set))

    def _get_c12n_groups(self, c12n: str, long_format: bool = True) -> Tuple[List, List]:
        # Parse classifications in uppercase mode only
        c12n = c12n.upper()

        g1_set = set()
        g2_set = set()
        others = set()

        grp_part = c12n.split("//")
        groups = []
        for gp in grp_part:
            gp = gp.replace("REL TO ", "")
            temp_group = set([x.strip() for x in gp.split(",")])
            for t in temp_group:
                groups.extend(t.split("/"))

        for g in groups:
            if g in self.groups_map_lts:
                g1_set.add(self.groups_map_lts[g])
            elif g in self.groups_map_stl:
                g1_set.add(g)
            elif g in self.groups_aliases:
                for a in self.groups_aliases[g]:
                    g1_set.add(a)
            elif g in self.subgroups_map_lts:
                g2_set.add(self.subgroups_map_lts[g])
            elif g in self.subgroups_map_stl:
                g2_set.add(g)
            elif g in self.subgroups_aliases:
                for a in self.subgroups_aliases[g]:
                    g2_set.add(a)
            else:
                others.add(g)

        if self.dynamic_groups:
            for o in others:
                if o not in self.access_req_map_lts \
                        and o not in self.access_req_map_stl \
                        and o not in self.access_req_aliases \
                        and o not in self.levels_map \
                        and o not in self.levels_map_lts \
                        and o not in self.levels_aliases:
                    g1_set.add(o)

        if long_format:
            return sorted(
                [self.groups_map_stl.get(r, r) for r in g1_set]), sorted(
                [self.subgroups_map_stl[r] for r in g2_set])
        return sorted(list(g1_set)), sorted(list(g2_set))

    @staticmethod
    def _can_see_required(user_req: List, req: List) -> bool:
        return set(req).issubset(user_req)

    @staticmethod
    def _can_see_groups(user_groups: List, req: List) -> bool:
        if len(req) == 0:
            return True

        for g in user_groups:
            if g in req:
                return True

        return False

    # noinspection PyTypeChecker
    def _get_normalized_classification_text(self, lvl_idx: int, req: List, groups: List, subgroups: List,
                                            long_format: bool = True, skip_auto_select: bool = False) -> str:
        # 1. Check for all required items if they need a specific classification lvl
        required_lvl_idx = 0
        for r in req:
            required_lvl_idx = max(required_lvl_idx, self.params_map.get(r, {}).get("require_lvl", 0))
        out = self._get_c12n_level_text(max(lvl_idx, required_lvl_idx), long_format=long_format)

        # 2. Check for all required items if they should be shown inside the groups display part
        req_grp = []
        for r in req:
            if self.params_map.get(r, {}).get('is_required_group'):
                req_grp.append(r)
        req = list(set(req).difference(set(req_grp)))

        if req:
            out += "//" + "/".join(req)
        if req_grp:
            out += "//" + "/".join(sorted(req_grp))

        # 3. Add auto-selected subgroups
        if long_format:
            if len(subgroups) > 0 and len(self.subgroups_auto_select) > 0 and not skip_auto_select:
                subgroups = sorted(list(set(subgroups).union(set(self.subgroups_auto_select))))
        else:
            if len(subgroups) > 0 and len(self.subgroups_auto_select_short) > 0 and not skip_auto_select:
                subgroups = sorted(list(set(subgroups).union(set(self.subgroups_auto_select_short))))

        # 4. For every subgroup, check if the subgroup requires or is limited to a specific group
        temp_groups = []
        for sg in subgroups:
            required_group = self.params_map.get(sg, {}).get("require_group", None)
            if required_group is not None:
                temp_groups.append(required_group)

            limited_to_group = self.params_map.get(sg, {}).get("limited_to_group", None)
            if limited_to_group is not None:
                if limited_to_group in temp_groups:
                    temp_groups = [limited_to_group]
                else:
                    temp_groups = []

        for g in temp_groups:
            if long_format:
                groups.append(self.groups_map_stl.get(g, g))
            else:
                groups.append(self.groups_map_lts.get(g, g))
        groups = list(set(groups))

        # 5. Add auto-selected groups
        if long_format:
            if len(groups) > 0 and len(self.groups_auto_select) > 0 and not skip_auto_select:
                groups = sorted(list(set(groups).union(set(self.groups_auto_select))))
        else:
            if len(groups) > 0 and len(self.groups_auto_select_short) > 0 and not skip_auto_select:
                groups = sorted(list(set(groups).union(set(self.groups_auto_select_short))))

        if groups:
            out += {True: "/", False: "//"}[len(req_grp) > 0]
            if len(groups) == 1:
                # 6. If only one group, check if it has a solitary display name.
                grp = groups[0]
                display_name = self.params_map.get(grp, {}).get('solitary_display_name', grp)
                if display_name != grp:
                    out += display_name
                else:
                    out += "REL TO " + grp
            else:
                if not long_format:
                    # 7. In short format mode, check if there is an alias that can replace multiple groups
                    for alias, values in self.groups_aliases.items():
                        if len(values) > 1:
                            if sorted(values) == groups:
                                groups = [alias]
                out += "REL TO " + ", ".join(sorted(groups))

        if subgroups:
            if len(groups) > 0 or len(req_grp) > 0:
                out += "/"
            else:
                out += "//"
            out += "/".join(sorted(subgroups))

        return out

    def _get_classification_parts(self, c12n: str, long_format: bool = True) \
            -> Tuple[Union[Union[int, str], Any], List, List, List]:
        lvl_idx = self._get_c12n_level_index(c12n)
        req = self._get_c12n_required(c12n, long_format=long_format)
        groups, subgroups = self._get_c12n_groups(c12n, long_format=long_format)

        return lvl_idx, req, groups, subgroups

    @staticmethod
    def _max_groups(groups_1: List, groups_2: List) -> List:
        if len(groups_1) > 0 and len(groups_2) > 0:
            groups = set(groups_1) & set(groups_2)
        else:
            groups = set(groups_1) | set(groups_2)

        if len(groups_1) > 0 and len(groups_2) > 0 and len(groups) == 0:
            # NOTE: Intersection generated nothing, we will raise an InvalidClassification exception
            raise InvalidClassification("Could not find any intersection between the groups. %s & %s" % (groups_1,
                                                                                                         groups_2))

        return list(groups)

    # ++++++++++++++++++++++++
    # Public functions
    # ++++++++++++++++++++++++
    # noinspection PyUnusedLocal
    def list_all_classification_combinations(self, long_format: bool = True) -> Set:
        combinations = set()

        levels = self._list_items_and_aliases(self.original_definition['levels'], long_format=long_format)
        reqs = self._list_items_and_aliases(self.original_definition['required'], long_format=long_format)
        grps = self._list_items_and_aliases(self.original_definition['groups'], long_format=long_format)
        sgrps = self._list_items_and_aliases(self.original_definition['subgroups'], long_format=long_format)

        req_cbs = self._build_combinations(reqs)
        if long_format:
            grp_solitary_display = {
                x['name']: x['solitary_display_name'] for x in self.original_definition['groups']
                if 'solitary_display_name' in x
            }
        else:
            grp_solitary_display = {
                x['short_name']: x['solitary_display_name'] for x in self.original_definition['groups']
                if 'solitary_display_name' in x
            }
        solitary_names = [x['solitary_display_name'] for x in self.original_definition['groups']
                          if 'solitary_display_name' in x]

        grp_cbs = self._build_combinations(grps, separator=", ", solitary_display=grp_solitary_display)
        sgrp_cbs = self._build_combinations(sgrps)

        for p in itertools.product(levels, req_cbs):
            cl = "//".join(p)
            if cl.endswith("//"):
                combinations.add(cl[:-2])
            else:
                combinations.add(cl)

        temp_combinations = copy(combinations)
        for p in itertools.product(temp_combinations, grp_cbs):
            cl = "//REL TO ".join(p)
            if cl.endswith("//REL TO "):
                combinations.add(cl[:-9])
            else:
                combinations.add(cl)

        for sol_name in solitary_names:
            to_edit = []
            to_find = "REL TO {sol_name}".format(sol_name=sol_name)
            for c in combinations:
                if to_find in c:
                    to_edit.append(c)

            for e in to_edit:
                combinations.add(e.replace(to_find, sol_name))
                combinations.remove(e)

        temp_combinations = copy(combinations)
        for p in itertools.product(temp_combinations, sgrp_cbs):
            if "//REL TO " in p[0]:
                cl = "/".join(p)

                if cl.endswith("/"):
                    combinations.add(cl[:-1])
                else:
                    combinations.add(cl)
            else:
                cl = "//REL TO ".join(p)

                if cl.endswith("//REL TO "):
                    combinations.add(cl[:-9])
                else:
                    combinations.add(cl)

        return combinations

    # noinspection PyUnusedLocal
    def default_user_classification(self, user: Optional[str] = None, long_format: bool = True) -> str:
        """
        You can overload this function to specify a way to get the default classification of a user.
        By default, this function returns the UNRESTRICTED value of your classification definition.

        Args:
            user: Which user to get the classification for
            long_format: Request a long classification format or not

        Returns:
            The classification in the specified format
        """
        return self.UNRESTRICTED

    def get_parsed_classification_definition(self) -> Dict:
        """
        Returns all dictionary of all the variables inside the classification object that will be used
        to enforce classification throughout the system.
        """
        from copy import deepcopy
        out = deepcopy(self.__dict__)
        out['levels_map'].pop("INV", None)
        out['levels_map'].pop(str(self.INVALID_LVL), None)
        out['levels_map_stl'].pop("INV", None)
        out['levels_map_lts'].pop("INVALID", None)
        out['levels_map'].pop("NULL", None)
        out['levels_map'].pop(str(self.NULL_LVL), None)
        out['levels_map_stl'].pop("NULL", None)
        out['levels_map_lts'].pop("NULL", None)
        out.pop('_classification_cache', None)
        out.pop('_classification_cache_short', None)
        return out

    def get_access_control_parts(self, c12n: str, user_classification: bool = False) -> Dict:
        """
        Returns a dictionary containing the different access parameters Lucene needs to build it's queries

        Args:
            c12n: The classification to get the parts from
            user_classification: Is a user classification
        """
        if not self.enforce or self.invalid_mode:
            c12n = self.UNRESTRICTED

        try:
            # Normalize the classification before gathering the parts
            c12n = self.normalize_classification(c12n, skip_auto_select=user_classification)

            access_lvl = self._get_c12n_level_index(c12n)
            access_req = self._get_c12n_required(c12n, long_format=False)
            access_grp1, access_grp2 = self._get_c12n_groups(c12n, long_format=False)

            return {
                '__access_lvl__': access_lvl,
                '__access_req__': access_req,
                '__access_grp1__': access_grp1 or ['__EMPTY__'],
                '__access_grp2__': access_grp2 or ['__EMPTY__']
            }
        except InvalidClassification:
            if not self.enforce or self.invalid_mode:
                return {
                    '__access_lvl__': self.NULL_LVL,
                    '__access_req__': [],
                    '__access_grp1__': ['__EMPTY__'],
                    '__access_grp2__': ['__EMPTY__']
                }
            else:
                raise

    def get_access_control_req(self) -> Union[KeysView, List]:
        """
        Returns a list of the different possible REQUIRED parts
        """
        if not self.enforce or self.invalid_mode:
            return []

        return self.access_req_map_stl.keys()

    def get_access_control_groups(self) -> Union[KeysView, List]:
        """
        Returns a list of the different possible GROUPS
        """
        if not self.enforce or self.invalid_mode:
            return []

        return self.groups_map_stl.keys()

    def get_access_control_subgroups(self) -> Union[KeysView, List]:
        """
        Returns a list of the different possible SUBGROUPS
        """
        if not self.enforce or self.invalid_mode:
            return []

        return self.subgroups_map_stl.keys()

    def intersect_user_classification(self, user_c12n_1: str, user_c12n_2: str, long_format: bool = True) -> str:
        """
        This function intersects two user classification to return the maximum classification
        that both user could see.

        Args:
            user_c12n_1: First user classification
            user_c12n_2: Second user classification
            long_format: True/False in long format

        Returns:
            Intersected classification in the desired format
        """
        if not self.enforce or self.invalid_mode:
            return self.UNRESTRICTED

        # Normalize classifications before comparing them
        if user_c12n_1 is not None:
            user_c12n_1 = self.normalize_classification(user_c12n_1, skip_auto_select=True)
        if user_c12n_2 is not None:
            user_c12n_2 = self.normalize_classification(user_c12n_2, skip_auto_select=True)

        if user_c12n_1 is None:
            return user_c12n_2
        if user_c12n_2 is None:
            return user_c12n_1

        lvl_idx_1, req_1, groups_1, subgroups_1 = self._get_classification_parts(user_c12n_1, long_format=long_format)
        lvl_idx_2, req_2, groups_2, subgroups_2 = self._get_classification_parts(user_c12n_2, long_format=long_format)

        req = list(set(req_1) & set(req_2))
        groups = list(set(groups_1) & set(groups_2))
        subgroups = list(set(subgroups_1) & set(subgroups_2))

        return self._get_normalized_classification_text(min(lvl_idx_1, lvl_idx_2),
                                                        req,
                                                        groups,
                                                        subgroups,
                                                        long_format=long_format,
                                                        skip_auto_select=True)

    def is_accessible(self, user_c12n: str, c12n: str) -> bool:
        """
        Given a user classification, check if a user is allow to see a certain classification

        Args:
            user_c12n: Maximum classification for the user
            c12n: Classification the user which to see

        Returns:
            True is the user can see the classification
        """
        if self.invalid_mode:
            return False

        if not self.enforce:
            return True

        if c12n is None:
            return True

        # Normalize classifications before comparing them
        user_c12n = self.normalize_classification(user_c12n, skip_auto_select=True)
        c12n = self.normalize_classification(c12n, skip_auto_select=True)

        user_req = self._get_c12n_required(user_c12n)
        user_groups, user_subgroups = self._get_c12n_groups(user_c12n)
        req = self._get_c12n_required(c12n)
        groups, subgroups = self._get_c12n_groups(c12n)

        if self._get_c12n_level_index(user_c12n) >= self._get_c12n_level_index(c12n):
            if not self._can_see_required(user_req, req):
                return False
            if not self._can_see_groups(user_groups, groups):
                return False
            if not self._can_see_groups(user_subgroups, subgroups):
                return False
            return True
        return False

    def is_valid(self, c12n: str, skip_auto_select: bool = False) -> bool:
        """
        Performs a series of checks againts a classification to make sure it is valid in it's current form

        Args:
            c12n: The classification we want to validate
            skip_auto_select: skip the auto selection phase

        Returns:
            True if the classification is valid
        """
        if not self.enforce:
            return True

        try:
            # Classification normalization test
            n_c12n = self.normalize_classification(c12n, skip_auto_select=skip_auto_select)
            n_lvl_idx, n_req, n_groups, n_subgroups = self._get_classification_parts(n_c12n)
            lvl_idx, req, groups, subgroups = self._get_classification_parts(c12n)
        except InvalidClassification:
            return False

        if lvl_idx != n_lvl_idx:
            return False

        if sorted(req) != sorted(n_req):
            return False

        if sorted(groups) != sorted(n_groups):
            return False

        if sorted(subgroups) != sorted(n_subgroups):
            return False

        c12n = c12n.replace("REL TO ", "")
        parts = c12n.split("//")

        # There is a maximum of 3 parts
        if len(parts) > 3:
            return False

        cur_part = parts.pop(0)
        # First parts as to be a classification level part
        if cur_part not in self.levels_aliases.keys() and \
                cur_part not in self.levels_map_lts.keys() and \
                cur_part not in self.levels_map_stl.keys():
            return False

        check_groups = False
        while len(parts) > 0:
            # Can't be two groups sections.
            if check_groups:
                return False

            cur_part = parts.pop(0)
            items = cur_part.split("/")
            comma_idx = None
            for idx, i in enumerate(items):
                if "," in i:
                    comma_idx = idx

            if comma_idx is not None:
                items += [x.strip() for x in items.pop(comma_idx).split(",")]

            for i in items:
                if not check_groups:
                    # If current item not found in access req, we might already be dealing with groups
                    if i not in self.access_req_aliases.keys() and \
                            i not in self.access_req_map_stl.keys() and \
                            i not in self.access_req_map_lts.keys():
                        check_groups = True

                if check_groups and not self.dynamic_groups:
                    # If not groups. That stuff does not exists...
                    if i not in self.groups_aliases.keys() and \
                            i not in self.groups_map_stl.keys() and \
                            i not in self.groups_map_lts.keys() and \
                            i not in self.subgroups_aliases.keys() and \
                            i not in self.subgroups_map_stl.keys() and \
                            i not in self.subgroups_map_lts.keys():
                        return False

        return True

    def max_classification(self, c12n_1: str, c12n_2: str, long_format: bool = True) -> str:
        """
        Mixes to classification and returns to most restrictive form for them

        Args:
            c12n_1: First classification
            c12n_2: Second classification
            long_format: True/False in long format

        Returns:
            The most restrictive classification that we could create out of the two
        """
        if not self.enforce or self.invalid_mode:
            return self.UNRESTRICTED

        # Normalize classifications before comparing them
        if c12n_1 is not None:
            c12n_1 = self.normalize_classification(c12n_1)
        if c12n_2 is not None:
            c12n_2 = self.normalize_classification(c12n_2)

        if c12n_1 is None:
            return c12n_2
        if c12n_2 is None:
            return c12n_1

        lvl_idx_1, req_1, groups_1, subgroups_1 = self._get_classification_parts(c12n_1, long_format=long_format)
        lvl_idx_2, req_2, groups_2, subgroups_2 = self._get_classification_parts(c12n_2, long_format=long_format)

        req = list(set(req_1) | set(req_2))
        groups = self._max_groups(groups_1, groups_2)
        subgroups = self._max_groups(subgroups_1, subgroups_2)

        return self._get_normalized_classification_text(max(lvl_idx_1, lvl_idx_2),
                                                        req,
                                                        groups,
                                                        subgroups,
                                                        long_format=long_format)

    def min_classification(self, c12n_1: str, c12n_2: str, long_format: bool = True) -> str:
        """
        Mixes to classification and returns to least restrictive form for them

        Args:
            c12n_1: First classification
            c12n_2: Second classification
            long_format: True/False in long format

        Returns:
            The least restrictive classification that we could create out of the two
        """
        if not self.enforce or self.invalid_mode:
            return self.UNRESTRICTED

        # Normalize classifications before comparing them
        if c12n_1 is not None:
            c12n_1 = self.normalize_classification(c12n_1)
        if c12n_2 is not None:
            c12n_2 = self.normalize_classification(c12n_2)

        if c12n_1 is None:
            return c12n_2
        if c12n_2 is None:
            return c12n_1

        lvl_idx_1, req_1, groups_1, subgroups_1 = self._get_classification_parts(c12n_1, long_format=long_format)
        lvl_idx_2, req_2, groups_2, subgroups_2 = self._get_classification_parts(c12n_2, long_format=long_format)

        req = list(set(req_1) & set(req_2))
        if len(groups_1) > 0 and len(groups_2) > 0:
            groups = list(set(groups_1) | set(groups_2))
        else:
            groups = []

        if len(subgroups_1) > 0 and len(subgroups_2) > 0:
            subgroups = list(set(subgroups_1) | set(subgroups_2))
        else:
            subgroups = []

        return self._get_normalized_classification_text(min(lvl_idx_1, lvl_idx_2),
                                                        req,
                                                        groups,
                                                        subgroups,
                                                        long_format=long_format)

    def normalize_classification(self, c12n: str, long_format: bool = True, skip_auto_select: bool = False) -> str:
        """
        Normalize a given classification by applying the rules defined in the classification definition.
        This function will remove any invalid parts and add missing parts to the classification.
        It will also ensure that the display of the classification is always done the same way

        Args:
            c12n: Classification to normalize
            long_format: True/False in long format
            skip_auto_select: True/False skip group auto adding, use True when dealing with user's classifications

        Returns:
            A normalized version of the original classification
        """
        if not self.enforce or self.invalid_mode:
            return self.UNRESTRICTED

        # Has the classification has already been normalized before?
        if long_format and c12n in self._classification_cache:
            return c12n
        if not long_format and c12n in self._classification_cache_short:
            return c12n

        lvl_idx, req, groups, subgroups = self._get_classification_parts(c12n, long_format=long_format)
        new_c12n = self._get_normalized_classification_text(lvl_idx, req, groups, subgroups,
                                                            long_format=long_format,
                                                            skip_auto_select=skip_auto_select)
        if long_format:
            self._classification_cache.add(new_c12n)
        else:
            self._classification_cache_short.add(new_c12n)

        return new_c12n

    def build_user_classification(self, c12n_1: str, c12n_2: str, long_format: bool = True) -> str:
        """
        Mixes to classification and return the classification marking that would give access to the most data

        Args:
            c12n_1: First classification
            c12n_2: Second classification
            long_format: True/False in long format

        Returns:
            The classification that would give access to the most data
        """
        if not self.enforce or self.invalid_mode:
            return self.UNRESTRICTED

        # Normalize classifications before comparing them
        if c12n_1 is not None:
            c12n_1 = self.normalize_classification(c12n_1, skip_auto_select=True)
        if c12n_2 is not None:
            c12n_2 = self.normalize_classification(c12n_2, skip_auto_select=True)

        if c12n_1 is None:
            return c12n_2
        if c12n_2 is None:
            return c12n_1

        lvl_idx_1, req_1, groups_1, subgroups_1 = self._get_classification_parts(c12n_1, long_format=long_format)
        lvl_idx_2, req_2, groups_2, subgroups_2 = self._get_classification_parts(c12n_2, long_format=long_format)

        req = list(set(req_1) | set(req_2))
        groups = list(set(groups_1) | set(groups_2))
        subgroups = list(set(subgroups_1) | set(subgroups_2))

        return self._get_normalized_classification_text(max(lvl_idx_1, lvl_idx_2), req, groups, subgroups,
                                                        long_format=long_format, skip_auto_select=True)



from assemblyline import odm
from assemblyline.common import forge
from assemblyline.odm.models.submission import ServiceSelection

Classification = forge.get_classification()

ENCODINGS = {"cart", "raw"}
VIEWS = {"report", "details"}


@odm.model(index=False, store=False)
class UserSettings(odm.Model):                                      # User's default settings
    classification = odm.Classification(
        default=Classification.UNRESTRICTED)                          # Default submission classification
    deep_scan = odm.Boolean(default=False)                            # Should a deep scan be performed
    description = odm.Keyword(default="")                             # Default description
    download_encoding = odm.Enum(values=ENCODINGS, default="cart")    # Default download encoding when downloading files
    expand_min_score = odm.Integer(default=500)                       # Auto-expand section when score bigger then this
    ignore_cache = odm.Boolean(default=False)                         # Ignore service caching
    ignore_dynamic_recursion_prevention = odm.Boolean(default=False)  # Ignore dynamic recursion prevention
    ignore_filtering = odm.Boolean(default=False)                     # Ignore filtering services
    malicious = odm.Boolean(default=False)                            # Is the file submitted known to be malicious
    priority = odm.Integer(default=1000)                              # Default priority for the submissions
    profile = odm.Boolean(default=False)                              # Should the submission do extra profiling
    service_spec = odm.Mapping(odm.Any(), default={})                 # Default service specific settings
    services = odm.Compound(ServiceSelection, default={})             # Default service selection
    submission_view = odm.Enum(values=VIEWS, default="report")        # Default view for completed submissions
    ttl = odm.Integer(default=30)                                     # Default submission Time to Live (days)



# from https://data.iana.org/TLD/tlds-alpha-by-domain.txt
# Version 2018051800, Last Updated Fri May 18 07:07:02 2018 UTC
TLDS_ALPHA_BY_DOMAIN = [
    "AAA",
    "AARP",
    "ABARTH",
    "ABB",
    "ABBOTT",
    "ABBVIE",
    "ABC",
    "ABLE",
    "ABOGADO",
    "ABUDHABI",
    "AC",
    "ACADEMY",
    "ACCENTURE",
    "ACCOUNTANT",
    "ACCOUNTANTS",
    "ACO",
    "ACTIVE",
    "ACTOR",
    "AD",
    "ADAC",
    "ADS",
    "ADULT",
    "AE",
    "AEG",
    "AERO",
    "AETNA",
    "AF",
    "AFAMILYCOMPANY",
    "AFL",
    "AFRICA",
    "AG",
    "AGAKHAN",
    "AGENCY",
    "AI",
    "AIG",
    "AIGO",
    "AIRBUS",
    "AIRFORCE",
    "AIRTEL",
    "AKDN",
    "AL",
    "ALFAROMEO",
    "ALIBABA",
    "ALIPAY",
    "ALLFINANZ",
    "ALLSTATE",
    "ALLY",
    "ALSACE",
    "ALSTOM",
    "AM",
    "AMERICANEXPRESS",
    "AMERICANFAMILY",
    "AMEX",
    "AMFAM",
    "AMICA",
    "AMSTERDAM",
    "ANALYTICS",
    "ANDROID",
    "ANQUAN",
    "ANZ",
    "AO",
    "AOL",
    "APARTMENTS",
    "APP",
    "APPLE",
    "AQ",
    "AQUARELLE",
    "AR",
    "ARAB",
    "ARAMCO",
    "ARCHI",
    "ARMY",
    "ARPA",
    "ART",
    "ARTE",
    "AS",
    "ASDA",
    "ASIA",
    "ASSOCIATES",
    "AT",
    "ATHLETA",
    "ATTORNEY",
    "AU",
    "AUCTION",
    "AUDI",
    "AUDIBLE",
    "AUDIO",
    "AUSPOST",
    "AUTHOR",
    "AUTO",
    "AUTOS",
    "AVIANCA",
    "AW",
    "AWS",
    "AX",
    "AXA",
    "AZ",
    "AZURE",
    "BA",
    "BABY",
    "BAIDU",
    "BANAMEX",
    "BANANAREPUBLIC",
    "BAND",
    "BANK",
    "BAR",
    "BARCELONA",
    "BARCLAYCARD",
    "BARCLAYS",
    "BAREFOOT",
    "BARGAINS",
    "BASEBALL",
    "BASKETBALL",
    "BAUHAUS",
    "BAYERN",
    "BB",
    "BBC",
    "BBT",
    "BBVA",
    "BCG",
    "BCN",
    "BD",
    "BE",
    "BEATS",
    "BEAUTY",
    "BEER",
    "BENTLEY",
    "BERLIN",
    "BEST",
    "BESTBUY",
    "BET",
    "BF",
    "BG",
    "BH",
    "BHARTI",
    "BI",
    "BIBLE",
    "BID",
    "BIKE",
    "BING",
    "BINGO",
    "BIO",
    "BIZ",
    "BJ",
    "BLACK",
    "BLACKFRIDAY",
    "BLANCO",
    "BLOCKBUSTER",
    "BLOG",
    "BLOOMBERG",
    "BLUE",
    "BM",
    "BMS",
    "BMW",
    "BN",
    "BNL",
    "BNPPARIBAS",
    "BO",
    "BOATS",
    "BOEHRINGER",
    "BOFA",
    "BOM",
    "BOND",
    "BOO",
    "BOOK",
    "BOOKING",
    "BOSCH",
    "BOSTIK",
    "BOSTON",
    "BOT",
    "BOUTIQUE",
    "BOX",
    "BR",
    "BRADESCO",
    "BRIDGESTONE",
    "BROADWAY",
    "BROKER",
    "BROTHER",
    "BRUSSELS",
    "BS",
    "BT",
    "BUDAPEST",
    "BUGATTI",
    "BUILD",
    "BUILDERS",
    "BUSINESS",
    "BUY",
    "BUZZ",
    "BV",
    "BW",
    "BY",
    "BZ",
    "BZH",
    "CA",
    "CAB",
    "CAFE",
    "CAL",
    "CALL",
    "CALVINKLEIN",
    "CAM",
    "CAMERA",
    "CAMP",
    "CANCERRESEARCH",
    "CANON",
    "CAPETOWN",
    "CAPITAL",
    "CAPITALONE",
    "CAR",
    "CARAVAN",
    "CARDS",
    "CARE",
    "CAREER",
    "CAREERS",
    "CARS",
    "CARTIER",
    "CASA",
    "CASE",
    "CASEIH",
    "CASH",
    "CASINO",
    "CAT",
    "CATERING",
    "CATHOLIC",
    "CBA",
    "CBN",
    "CBRE",
    "CBS",
    "CC",
    "CD",
    "CEB",
    "CENTER",
    "CEO",
    "CERN",
    "CF",
    "CFA",
    "CFD",
    "CG",
    "CH",
    "CHANEL",
    "CHANNEL",
    "CHARITY",
    "CHASE",
    "CHAT",
    "CHEAP",
    "CHINTAI",
    "CHRISTMAS",
    "CHROME",
    "CHRYSLER",
    "CHURCH",
    "CI",
    "CIPRIANI",
    "CIRCLE",
    "CISCO",
    "CITADEL",
    "CITI",
    "CITIC",
    "CITY",
    "CITYEATS",
    "CK",
    "CL",
    "CLAIMS",
    "CLEANING",
    "CLICK",
    "CLINIC",
    "CLINIQUE",
    "CLOTHING",
    "CLOUD",
    "CLUB",
    "CLUBMED",
    "CM",
    "CN",
    "CO",
    "COACH",
    "CODES",
    "COFFEE",
    "COLLEGE",
    "COLOGNE",
    "COM",
    "COMCAST",
    "COMMBANK",
    "COMMUNITY",
    "COMPANY",
    "COMPARE",
    "COMPUTER",
    "COMSEC",
    "CONDOS",
    "CONSTRUCTION",
    "CONSULTING",
    "CONTACT",
    "CONTRACTORS",
    "COOKING",
    "COOKINGCHANNEL",
    "COOL",
    "COOP",
    "CORSICA",
    "COUNTRY",
    "COUPON",
    "COUPONS",
    "COURSES",
    "CR",
    "CREDIT",
    "CREDITCARD",
    "CREDITUNION",
    "CRICKET",
    "CROWN",
    "CRS",
    "CRUISE",
    "CRUISES",
    "CSC",
    "CU",
    "CUISINELLA",
    "CV",
    "CW",
    "CX",
    "CY",
    "CYMRU",
    "CYOU",
    "CZ",
    "DABUR",
    "DAD",
    "DANCE",
    "DATA",
    "DATE",
    "DATING",
    "DATSUN",
    "DAY",
    "DCLK",
    "DDS",
    "DE",
    "DEAL",
    "DEALER",
    "DEALS",
    "DEGREE",
    "DELIVERY",
    "DELL",
    "DELOITTE",
    "DELTA",
    "DEMOCRAT",
    "DENTAL",
    "DENTIST",
    "DESI",
    "DESIGN",
    "DEV",
    "DHL",
    "DIAMONDS",
    "DIET",
    "DIGITAL",
    "DIRECT",
    "DIRECTORY",
    "DISCOUNT",
    "DISCOVER",
    "DISH",
    "DIY",
    "DJ",
    "DK",
    "DM",
    "DNP",
    "DO",
    "DOCS",
    "DOCTOR",
    "DODGE",
    "DOG",
    "DOHA",
    "DOMAINS",
    "DOT",
    "DOWNLOAD",
    "DRIVE",
    "DTV",
    "DUBAI",
    "DUCK",
    "DUNLOP",
    "DUNS",
    "DUPONT",
    "DURBAN",
    "DVAG",
    "DVR",
    "DZ",
    "EARTH",
    "EAT",
    "EC",
    "ECO",
    "EDEKA",
    "EDU",
    "EDUCATION",
    "EE",
    "EG",
    "EMAIL",
    "EMERCK",
    "ENERGY",
    "ENGINEER",
    "ENGINEERING",
    "ENTERPRISES",
    "EPOST",
    "EPSON",
    "EQUIPMENT",
    "ER",
    "ERICSSON",
    "ERNI",
    "ES",
    "ESQ",
    "ESTATE",
    "ESURANCE",
    "ET",
    "ETISALAT",
    "EU",
    "EUROVISION",
    "EUS",
    "EVENTS",
    "EVERBANK",
    "EXCHANGE",
    "EXPERT",
    "EXPOSED",
    "EXPRESS",
    "EXTRASPACE",
    "FAGE",
    "FAIL",
    "FAIRWINDS",
    "FAITH",
    "FAMILY",
    "FAN",
    "FANS",
    "FARM",
    "FARMERS",
    "FASHION",
    "FAST",
    "FEDEX",
    "FEEDBACK",
    "FERRARI",
    "FERRERO",
    "FI",
    "FIAT",
    "FIDELITY",
    "FIDO",
    "FILM",
    "FINAL",
    "FINANCE",
    "FINANCIAL",
    "FIRE",
    "FIRESTONE",
    "FIRMDALE",
    "FISH",
    "FISHING",
    "FIT",
    "FITNESS",
    "FJ",
    "FK",
    "FLICKR",
    "FLIGHTS",
    "FLIR",
    "FLORIST",
    "FLOWERS",
    "FLY",
    "FM",
    "FO",
    "FOO",
    "FOOD",
    "FOODNETWORK",
    "FOOTBALL",
    "FORD",
    "FOREX",
    "FORSALE",
    "FORUM",
    "FOUNDATION",
    "FOX",
    "FR",
    "FREE",
    "FRESENIUS",
    "FRL",
    "FROGANS",
    "FRONTDOOR",
    "FRONTIER",
    "FTR",
    "FUJITSU",
    "FUJIXEROX",
    "FUN",
    "FUND",
    "FURNITURE",
    "FUTBOL",
    "FYI",
    "GA",
    "GAL",
    "GALLERY",
    "GALLO",
    "GALLUP",
    "GAME",
    "GAMES",
    "GAP",
    "GARDEN",
    "GB",
    "GBIZ",
    "GD",
    "GDN",
    "GE",
    "GEA",
    "GENT",
    "GENTING",
    "GEORGE",
    "GF",
    "GG",
    "GGEE",
    "GH",
    "GI",
    "GIFT",
    "GIFTS",
    "GIVES",
    "GIVING",
    "GL",
    "GLADE",
    "GLASS",
    "GLE",
    "GLOBAL",
    "GLOBO",
    "GM",
    "GMAIL",
    "GMBH",
    "GMO",
    "GMX",
    "GN",
    "GODADDY",
    "GOLD",
    "GOLDPOINT",
    "GOLF",
    "GOO",
    "GOODHANDS",
    "GOODYEAR",
    "GOOG",
    "GOOGLE",
    "GOP",
    "GOT",
    "GOV",
    "GP",
    "GQ",
    "GR",
    "GRAINGER",
    "GRAPHICS",
    "GRATIS",
    "GREEN",
    "GRIPE",
    "GROCERY",
    "GROUP",
    "GS",
    "GT",
    "GU",
    "GUARDIAN",
    "GUCCI",
    "GUGE",
    "GUIDE",
    "GUITARS",
    "GURU",
    "GW",
    "GY",
    "HAIR",
    "HAMBURG",
    "HANGOUT",
    "HAUS",
    "HBO",
    "HDFC",
    "HDFCBANK",
    "HEALTH",
    "HEALTHCARE",
    "HELP",
    "HELSINKI",
    "HERE",
    "HERMES",
    "HGTV",
    "HIPHOP",
    "HISAMITSU",
    "HITACHI",
    "HIV",
    "HK",
    "HKT",
    "HM",
    "HN",
    "HOCKEY",
    "HOLDINGS",
    "HOLIDAY",
    "HOMEDEPOT",
    "HOMEGOODS",
    "HOMES",
    "HOMESENSE",
    "HONDA",
    "HONEYWELL",
    "HORSE",
    "HOSPITAL",
    "HOST",
    "HOSTING",
    "HOT",
    "HOTELES",
    "HOTELS",
    "HOTMAIL",
    "HOUSE",
    "HOW",
    "HR",
    "HSBC",
    "HT",
    "HU",
    "HUGHES",
    "HYATT",
    "HYUNDAI",
    "IBM",
    "ICBC",
    "ICE",
    "ICU",
    "ID",
    "IE",
    "IEEE",
    "IFM",
    "IKANO",
    "IL",
    "IM",
    "IMAMAT",
    "IMDB",
    "IMMO",
    "IMMOBILIEN",
    "IN",
    "INDUSTRIES",
    "INFINITI",
    "INFO",
    "INC",
    "ING",
    "INK",
    "INSTITUTE",
    "INSURANCE",
    "INSURE",
    "INT",
    "INTEL",
    "INTERNATIONAL",
    "INTUIT",
    "INVESTMENTS",
    "IO",
    "IPIRANGA",
    "IQ",
    "IR",
    "IRISH",
    "IS",
    "ISELECT",
    "ISMAILI",
    "IST",
    "ISTANBUL",
    "IT",
    "ITAU",
    "ITV",
    "IVECO",
    # "IWC",
    "JAGUAR",
    "JAVA",
    "JCB",
    "JCP",
    "JE",
    "JEEP",
    "JETZT",
    "JEWELRY",
    "JIO",
    "JLC",
    "JLL",
    "JM",
    "JMP",
    "JNJ",
    "JO",
    "JOBS",
    "JOBURG",
    "JOT",
    "JOY",
    "JP",
    "JPMORGAN",
    "JPRS",
    "JUEGOS",
    "JUNIPER",
    "KAUFEN",
    "KDDI",
    "KE",
    "KERRYHOTELS",
    "KERRYLOGISTICS",
    "KERRYPROPERTIES",
    "KFH",
    "KG",
    "KH",
    "KI",
    "KIA",
    "KIM",
    "KINDER",
    "KINDLE",
    "KITCHEN",
    "KIWI",
    "KM",
    "KN",
    "KOELN",
    "KOMATSU",
    "KOSHER",
    "KP",
    "KPMG",
    "KPN",
    "KR",
    "KRD",
    "KRED",
    "KUOKGROUP",
    "KW",
    "KY",
    "KYOTO",
    "KZ",
    "LA",
    "LACAIXA",
    "LADBROKES",
    "LAMBORGHINI",
    "LAMER",
    "LANCASTER",
    "LANCIA",
    "LANCOME",
    "LAND",
    "LANDROVER",
    "LANXESS",
    "LASALLE",
    "LAT",
    "LATINO",
    "LATROBE",
    "LAW",
    "LAWYER",
    "LB",
    "LC",
    "LDS",
    "LEASE",
    "LECLERC",
    "LEFRAK",
    "LEGAL",
    "LEGO",
    "LEXUS",
    "LGBT",
    "LI",
    "LIAISON",
    "LIDL",
    "LIFE",
    "LIFEINSURANCE",
    "LIFESTYLE",
    "LIGHTING",
    "LIKE",
    "LILLY",
    "LIMITED",
    "LIMO",
    "LINCOLN",
    "LINDE",
    "LINK",
    "LIPSY",
    "LIVE",
    "LIVING",
    "LIXIL",
    "LK",
    "LLC",
    "LOAN",
    "LOANS",
    "LOCKER",
    "LOCUS",
    "LOFT",
    "LOL",
    "LONDON",
    "LOTTE",
    "LOTTO",
    "LOVE",
    "LPL",
    "LPLFINANCIAL",
    "LR",
    "LS",
    "LT",
    "LTD",
    "LTDA",
    "LU",
    "LUNDBECK",
    "LUPIN",
    "LUXE",
    "LUXURY",
    "LV",
    "LY",
    "MA",
    "MACYS",
    "MADRID",
    "MAIF",
    "MAISON",
    "MAKEUP",
    "MAN",
    "MANAGEMENT",
    "MANGO",
    "MAP",
    "MARKET",
    "MARKETING",
    "MARKETS",
    "MARRIOTT",
    "MARSHALLS",
    "MASERATI",
    "MATTEL",
    "MBA",
    "MC",
    "MCKINSEY",
    "MD",
    "ME",
    "MED",
    "MEDIA",
    "MEET",
    "MELBOURNE",
    "MEME",
    "MEMORIAL",
    "MEN",
    "MENU",
    "MERCKMSD",
    "METLIFE",
    "MG",
    "MH",
    "MIAMI",
    "MICROSOFT",
    "MIL",
    "MINI",
    "MINT",
    "MIT",
    "MITSUBISHI",
    "MK",
    "ML",
    "MLB",
    "MLS",
    "MM",
    "MMA",
    "MN",
    "MO",
    "MOBI",
    "MOBILE",
    "MOBILY",
    "MODA",
    "MOE",
    "MOI",
    "MOM",
    "MONASH",
    "MONEY",
    "MONSTER",
    "MOPAR",
    "MORMON",
    "MORTGAGE",
    "MOSCOW",
    "MOTO",
    "MOTORCYCLES",
    "MOV",
    "MOVIE",
    "MOVISTAR",
    "MP",
    "MQ",
    "MR",
    "MS",
    "MSD",
    "MT",
    "MTN",
    "MTR",
    "MU",
    "MUSEUM",
    "MUTUAL",
    "MV",
    "MW",
    "MX",
    "MY",
    "MZ",
    "NA",
    "NAB",
    "NADEX",
    "NAGOYA",
    "NAME",
    "NATIONWIDE",
    "NATURA",
    "NAVY",
    "NBA",
    "NC",
    "NE",
    "NEC",
    "NET",
    "NETBANK",
    "NETFLIX",
    "NETWORK",
    "NEUSTAR",
    "NEW",
    "NEWHOLLAND",
    "NEWS",
    "NEXT",
    "NEXTDIRECT",
    "NEXUS",
    "NF",
    "NFL",
    "NG",
    "NGO",
    "NHK",
    "NI",
    "NICO",
    "NIKE",
    "NIKON",
    "NINJA",
    "NISSAN",
    "NISSAY",
    "NL",
    "NO",
    "NOKIA",
    "NORTHWESTERNMUTUAL",
    "NORTON",
    "NOW",
    "NOWRUZ",
    "NOWTV",
    "NP",
    "NR",
    "NRA",
    "NRW",
    "NTT",
    "NU",
    "NYC",
    "NZ",
    "OBI",
    "OBSERVER",
    "OFF",
    "OFFICE",
    "OKINAWA",
    "OLAYAN",
    "OLAYANGROUP",
    "OLDNAVY",
    "OLLO",
    "OM",
    "OMEGA",
    "ONE",
    "ONG",
    "ONL",
    "ONLINE",
    "ONYOURSIDE",
    "OOO",
    "OPEN",
    "ORACLE",
    "ORANGE",
    "ORG",
    "ORGANIC",
    "ORIGINS",
    "OSAKA",
    "OTSUKA",
    "OTT",
    "OVH",
    "PA",
    "PAGE",
    "PANASONIC",
    "PANERAI",
    "PARIS",
    "PARS",
    "PARTNERS",
    "PARTS",
    "PARTY",
    "PASSAGENS",
    "PAY",
    "PCCW",
    "PE",
    "PET",
    "PF",
    "PFIZER",
    "PG",
    "PH",
    "PHARMACY",
    "PHD",
    "PHILIPS",
    "PHONE",
    "PHOTO",
    "PHOTOGRAPHY",
    "PHOTOS",
    "PHYSIO",
    "PIAGET",
    "PICS",
    "PICTET",
    "PICTURES",
    "PID",
    "PIN",
    "PING",
    "PINK",
    "PIONEER",
    "PIZZA",
    "PK",
    "PL",
    "PLACE",
    "PLAY",
    "PLAYSTATION",
    "PLUMBING",
    "PLUS",
    "PM",
    "PN",
    "PNC",
    "POHL",
    "POKER",
    "POLITIE",
    "PORN",
    "POST",
    "PR",
    "PRAMERICA",
    "PRAXI",
    "PRESS",
    "PRIME",
    "PRO",
    "PROD",
    "PRODUCTIONS",
    "PROF",
    "PROGRESSIVE",
    "PROMO",
    "PROPERTIES",
    "PROPERTY",
    "PROTECTION",
    "PRU",
    "PRUDENTIAL",
    "PS",
    "PT",
    "PUB",
    "PW",
    "PWC",
    "PY",
    "QA",
    "QPON",
    "QUEBEC",
    "QUEST",
    "QVC",
    "RACING",
    "RADIO",
    "RAID",
    "RE",
    "READ",
    "REALESTATE",
    "REALTOR",
    "REALTY",
    "RECIPES",
    "RED",
    "REDSTONE",
    "REDUMBRELLA",
    "REHAB",
    "REISE",
    "REISEN",
    "REIT",
    "RELIANCE",
    "REN",
    "RENT",
    "RENTALS",
    "REPAIR",
    "REPORT",
    "REPUBLICAN",
    "REST",
    "RESTAURANT",
    "REVIEW",
    "REVIEWS",
    "REXROTH",
    "RICH",
    "RICHARDLI",
    "RICOH",
    "RIGHTATHOME",
    "RIL",
    "RIO",
    "RIP",
    "RMIT",
    "RO",
    "ROCHER",
    "ROCKS",
    "RODEO",
    "ROGERS",
    "ROOM",
    "RS",
    "RSVP",
    "RU",
    "RUGBY",
    "RUHR",
    "RUN",
    "RW",
    "RWE",
    "RYUKYU",
    "SA",
    "SAARLAND",
    "SAFE",
    "SAFETY",
    "SAKURA",
    "SALE",
    "SALON",
    "SAMSCLUB",
    "SAMSUNG",
    "SANDVIK",
    "SANDVIKCOROMANT",
    "SANOFI",
    "SAP",
    "SARL",
    "SAS",
    "SAVE",
    "SAXO",
    "SB",
    "SBI",
    "SBS",
    "SC",
    "SCA",
    "SCB",
    "SCHAEFFLER",
    "SCHMIDT",
    "SCHOLARSHIPS",
    "SCHOOL",
    "SCHULE",
    "SCHWARZ",
    "SCIENCE",
    "SCJOHNSON",
    "SCOR",
    "SCOT",
    "SD",
    "SE",
    "SEARCH",
    "SEAT",
    "SECURE",
    "SECURITY",
    "SEEK",
    "SELECT",
    "SENER",
    "SERVICES",
    "SES",
    "SEVEN",
    "SEW",
    "SEX",
    "SEXY",
    "SFR",
    "SG",
    "SH",
    "SHANGRILA",
    "SHARP",
    "SHAW",
    "SHELL",
    "SHIA",
    "SHIKSHA",
    "SHOES",
    "SHOP",
    "SHOPPING",
    "SHOUJI",
    "SHOW",
    "SHOWTIME",
    "SHRIRAM",
    "SI",
    "SILK",
    "SINA",
    "SINGLES",
    "SITE",
    "SJ",
    "SK",
    "SKI",
    "SKIN",
    "SKY",
    "SKYPE",
    "SL",
    "SLING",
    "SM",
    "SMART",
    "SMILE",
    "SN",
    "SNCF",
    "SO",
    "SOCCER",
    "SOCIAL",
    "SOFTBANK",
    "SOFTWARE",
    "SOHU",
    "SOLAR",
    "SOLUTIONS",
    "SONG",
    "SONY",
    "SOY",
    "SPACE",
    "SPIEGEL",
    "SPORT",
    "SPOT",
    "SPREADBETTING",
    "SR",
    "SRL",
    "SRT",
    "ST",
    "STADA",
    "STAPLES",
    "STAR",
    "STARHUB",
    "STATEBANK",
    "STATEFARM",
    "STATOIL",
    "STC",
    "STCGROUP",
    "STOCKHOLM",
    "STORAGE",
    "STORE",
    "STREAM",
    "STUDIO",
    "STUDY",
    "STYLE",
    "SU",
    "SUCKS",
    "SUPPLIES",
    "SUPPLY",
    "SUPPORT",
    "SURF",
    "SURGERY",
    "SUZUKI",
    "SV",
    "SWATCH",
    "SWIFTCOVER",
    "SWISS",
    "SX",
    "SY",
    "SYDNEY",
    "SYMANTEC",
    "SYSTEMS",
    "SZ",
    "TAB",
    "TAIPEI",
    "TALK",
    "TAOBAO",
    "TARGET",
    "TATAMOTORS",
    "TATAR",
    "TATTOO",
    "TAX",
    "TAXI",
    "TC",
    "TCI",
    "TD",
    "TDK",
    "TEAM",
    "TECH",
    "TECHNOLOGY",
    "TEL",
    "TELECITY",
    "TELEFONICA",
    "TEMASEK",
    "TENNIS",
    "TEVA",
    "TF",
    "TG",
    "TH",
    "THD",
    "THEATER",
    "THEATRE",
    "TIAA",
    "TICKETS",
    "TIENDA",
    "TIFFANY",
    "TIPS",
    "TIRES",
    "TIROL",
    "TJ",
    "TJMAXX",
    "TJX",
    "TK",
    "TKMAXX",
    "TL",
    "TM",
    "TMALL",
    "TN",
    "TO",
    "TODAY",
    "TOKYO",
    "TOOLS",
    "TOP",
    "TORAY",
    "TOSHIBA",
    "TOTAL",
    "TOURS",
    "TOWN",
    "TOYOTA",
    "TOYS",
    "TR",
    "TRADE",
    "TRADING",
    "TRAINING",
    "TRAVEL",
    "TRAVELCHANNEL",
    "TRAVELERS",
    "TRAVELERSINSURANCE",
    "TRUST",
    "TRV",
    "TT",
    "TUBE",
    "TUI",
    "TUNES",
    "TUSHU",
    "TV",
    "TVS",
    "TW",
    "TZ",
    "UA",
    "UBANK",
    "UBS",
    "UCONNECT",
    "UG",
    "UK",
    "UNICOM",
    "UNIVERSITY",
    "UNO",
    "UOL",
    "UPS",
    "US",
    "UY",
    "UZ",
    "VA",
    "VACATIONS",
    "VANA",
    "VANGUARD",
    "VC",
    "VE",
    "VEGAS",
    "VENTURES",
    "VERISIGN",
    "VERSICHERUNG",
    "VET",
    "VG",
    "VI",
    "VIAJES",
    "VIDEO",
    "VIG",
    "VIKING",
    "VILLAS",
    "VIN",
    "VIP",
    "VIRGIN",
    "VISA",
    "VISION",
    "VISTA",
    "VISTAPRINT",
    "VIVA",
    "VIVO",
    "VLAANDEREN",
    "VN",
    "VODKA",
    "VOLKSWAGEN",
    "VOLVO",
    "VOTE",
    "VOTING",
    "VOTO",
    "VOYAGE",
    "VU",
    "VUELOS",
    "WALES",
    "WALMART",
    "WALTER",
    "WANG",
    "WANGGOU",
    "WARMAN",
    "WATCH",
    "WATCHES",
    "WEATHER",
    "WEATHERCHANNEL",
    "WEBCAM",
    "WEBER",
    "WEBSITE",
    "WED",
    "WEDDING",
    "WEIBO",
    "WEIR",
    "WF",
    "WHOSWHO",
    "WIEN",
    "WIKI",
    "WILLIAMHILL",
    "WIN",
    "WINDOWS",
    "WINE",
    "WINNERS",
    "WME",
    "WOLTERSKLUWER",
    "WOODSIDE",
    "WORK",
    "WORKS",
    "WORLD",
    "WOW",
    "WS",
    "WTC",
    "WTF",
    "XBOX",
    "XEROX",
    "XFINITY",
    "XIHUAN",
    "XIN",
    "XN--11B4C3D",
    "XN--1CK2E1B",
    "XN--1QQW23A",
    "XN--2SCRJ9C",
    "XN--30RR7Y",
    "XN--3BST00M",
    "XN--3DS443G",
    "XN--3E0B707E",
    "XN--3HCRJ9C",
    "XN--3OQ18VL8PN36A",
    "XN--3PXU8K",
    "XN--42C2D9A",
    "XN--45BR5CYL",
    "XN--45BRJ9C",
    "XN--45Q11C",
    "XN--4GBRIM",
    "XN--54B7FTA0CC",
    "XN--55QW42G",
    "XN--55QX5D",
    "XN--5SU34J936BGSG",
    "XN--5TZM5G",
    "XN--6FRZ82G",
    "XN--6QQ986B3XL",
    "XN--80ADXHKS",
    "XN--80AO21A",
    "XN--80AQECDR1A",
    "XN--80ASEHDB",
    "XN--80ASWG",
    "XN--8Y0A063A",
    "XN--90A3AC",
    "XN--90AE",
    "XN--90AIS",
    "XN--9DBQ2A",
    "XN--9ET52U",
    "XN--9KRT00A",
    "XN--B4W605FERD",
    "XN--BCK1B9A5DRE4C",
    "XN--C1AVG",
    "XN--C2BR7G",
    "XN--CCK2B3B",
    "XN--CG4BKI",
    "XN--CLCHC0EA0B2G2A9GCD",
    "XN--CZR694B",
    "XN--CZRS0T",
    "XN--CZRU2D",
    "XN--D1ACJ3B",
    "XN--D1ALF",
    "XN--E1A4C",
    "XN--ECKVDTC9D",
    "XN--EFVY88H",
    "XN--ESTV75G",
    "XN--FCT429K",
    "XN--FHBEI",
    "XN--FIQ228C5HS",
    "XN--FIQ64B",
    "XN--FIQS8S",
    "XN--FIQZ9S",
    "XN--FJQ720A",
    "XN--FLW351E",
    "XN--FPCRJ9C3D",
    "XN--FZC2C9E2C",
    "XN--FZYS8D69UVGM",
    "XN--G2XX48C",
    "XN--GCKR3F0F",
    "XN--GECRJ9C",
    "XN--GK3AT1E",
    "XN--H2BREG3EVE",
    "XN--H2BRJ9C",
    "XN--H2BRJ9C8C",
    "XN--HXT814E",
    "XN--I1B6B1A6A2E",
    "XN--IMR513N",
    "XN--IO0A7I",
    "XN--J1AEF",
    "XN--J1AMH",
    "XN--J6W193G",
    "XN--JLQ61U9W7B",
    "XN--JVR189M",
    "XN--KCRX77D1X4A",
    "XN--KPRW13D",
    "XN--KPRY57D",
    "XN--KPU716F",
    "XN--KPUT3I",
    "XN--L1ACC",
    "XN--LGBBAT1AD8J",
    "XN--MGB9AWBF",
    "XN--MGBA3A3EJT",
    "XN--MGBA3A4F16A",
    "XN--MGBA7C0BBN0A",
    "XN--MGBAAKC7DVF",
    "XN--MGBAAM7A8H",
    "XN--MGBAB2BD",
    "XN--MGBAI9AZGQP6J",
    "XN--MGBAYH7GPA",
    "XN--MGBB9FBPOB",
    "XN--MGBBH1A",
    "XN--MGBBH1A71E",
    "XN--MGBC0A9AZCG",
    "XN--MGBCA7DZDO",
    "XN--MGBERP4A5D4AR",
    "XN--MGBGU82A",
    "XN--MGBI4ECEXP",
    "XN--MGBPL2FH",
    "XN--MGBT3DHD",
    "XN--MGBTX2B",
    "XN--MGBX4CD0AB",
    "XN--MIX891F",
    "XN--MK1BU44C",
    "XN--MXTQ1M",
    "XN--NGBC5AZD",
    "XN--NGBE9E0A",
    "XN--NGBRX",
    "XN--NODE",
    "XN--NQV7F",
    "XN--NQV7FS00EMA",
    "XN--NYQY26A",
    "XN--O3CW4H",
    "XN--OGBPF8FL",
    "XN--OTU796D",
    "XN--P1ACF",
    "XN--P1AI",
    "XN--PBT977C",
    "XN--PGBS0DH",
    "XN--PSSY2U",
    "XN--Q9JYB4C",
    "XN--QCKA1PMC",
    "XN--QXAM",
    "XN--RHQV96G",
    "XN--ROVU88B",
    "XN--RVC1E0AM3E",
    "XN--S9BRJ9C",
    "XN--SES554G",
    "XN--T60B56A",
    "XN--TCKWE",
    "XN--TIQ49XQYJ",
    "XN--UNUP4Y",
    "XN--VERMGENSBERATER-CTB",
    "XN--VERMGENSBERATUNG-PWB",
    "XN--VHQUV",
    "XN--VUQ861B",
    "XN--W4R85EL8FHU5DNRA",
    "XN--W4RS40L",
    "XN--WGBH1C",
    "XN--WGBL6A",
    "XN--XHQ521B",
    "XN--XKC2AL3HYE2A",
    "XN--XKC2DL3A5EE0H",
    "XN--Y9A3AQ",
    "XN--YFRO4I67O",
    "XN--YGBI2AMMX",
    "XN--ZFR164B",
    # "XPERIA",
    "XXX",
    "XYZ",
    "YACHTS",
    "YAHOO",
    "YAMAXUN",
    "YANDEX",
    "YE",
    "YODOBASHI",
    "YOGA",
    "YOKOHAMA",
    "YOU",
    "YOUTUBE",
    "YT",
    "YUN",
    "ZA",
    "ZAPPOS",
    "ZARA",
    "ZERO",
    "ZIP",
    "ZIPPO",
    "ZM",
    "ZONE",
    "ZUERICH",
    "ZW"]



from typing import Dict

from assemblyline.common.dict_utils import flatten

tag_map = {
    "attribution": {
        "actor": ['THREAT_ACTOR'],
        "campaign": ['CAMPAIGN_NAME'],
        "exploit": ['EXPLOIT_NAME'],
        "implant": ['IMPLANT_NAME'],
        "family": ['IMPLANT_FAMILY'],
        "network": ['NET_ATTRIBUTION'],
     },
    "av": {
        "heuristic": ['AV_HEURISTIC'],
        "virus_name": ['AV_VIRUS_NAME'],
    },
    "cert": {
        "extended_key_usage": ['CERT_EXTENDED_KEY_USAGE'],
        "issuer": ['ANDROID_CERT_ISSUER', 'CERT_ISSUER'],
        "key_usage": ['CERT_KEY_USAGE'],
        "owner": ['ANDROID_CERT_OWNER'],
        "serial_no": ['CERT_SERIAL_NO'],
        "signature_algo": ['CERT_SIGNATURE_ALGO'],
        "subject": ['CERT_SUBJECT'],
        "subject_alt_name": ['CERT_SUBJECT_ALT_NAME'],
        "thumbprint": ['CERT_THUMBPRINT'],
        "valid": {
            "start": ['ANDROID_CERT_START_DATE', 'CERT_VALID_FROM'],
            "end": ['ANDROID_CERT_END_DATE', 'CERT_VALID_TO'],
        },
        "version": ['CERT_VERSION'],
    },
    "dynamic": {
        "autorun_location": ['AUTORUN_LOCATION'],
        "dos_device": ['DOS_DEVICE_NAME,'],
        "mutex": ['DYNAMIC_MUTEX_NAME'],
        "registry_key": ['REGISTRY_KEY'],
        "process": {
            "file_name": ['DYNAMIC_PROCESS_FNAME'],
            "command_line": ['DYNAMIC_PROCESS_COMMANDLINE'],
        },
        "signature": {
            "category": ['DYNAMIC_SIGNATURE_CATEGORY'],
            "family": ['DYNAMIC_SIGNATURE_NAME'],
            "name": ['DYNAMIC_SIGNATURE_FAMILY'],
        },
        "ssdeep": {
            "cls_ids": ['DYNAMIC_CLSIDS_SSDEEP'],
            "regkeys": ['DYNAMIC_REGKEYS_SSDEEP'],
            "dynamic_classes": ['ANDROID_DYNAMIC_CLASSES_SSDEEP'],
        },
        "window": {
            "name": ['DYNAMIC_WINDOW_NAME'],
            "class_name": ['DYNAMIC_WINDOW_CLASSNAME'],
        },
    },
    "info": {
        "phone_number": ['NET_PHONE_NUMBER'],
    },
    "file": {
        "behavior": ['FILE_SUMMARY'],
        "compiler": ['INFO_COMPILER'],
        "config": ['FILE_CONFIG'],
        "date": {
            "creation": ['OLE_CREATION_TIME', 'PDF_DATE_CREATION'],
            "last_modified": ['OLE_LASTMOD_TIME', 'PDF_DATE_LASTMODIFIED']
        },
        "lib": ['INFO_LIBS'],
        "name": {
            "anomaly": ['FILENAME_ANOMALIES'],
            "extracted": ['FILE_NAME'],
        },
        "path": ['FILE_PATH_NAME', 'DYNAMIC_DROP_PATH,'],
        "rule": {
            "tagcheck": ['TAGCHECK_RULE'],
            "yara": ['FILE_YARA_RULE'],
        },
        "string": {
            "api": ['WIN_API_STRING'],
            "blacklisted": ['PESTUDIO_BLACKLIST_STRING'],
            "decoded": ['FILE_DECODED_STRING'],
            "extracted": ['FILE_STRING'],
        },
        "apk": {
            "activity": ['ANDROID_ACTIVITY'],
            "app": {
                "label": ['ANDROID_APP_LABEL'],
                "version": ['ANDROID_APP_VERSION'],
            },
            "feature": ['ANDROID_FEATURE'],
            "locale": ['ANDROID_LOCALE'],
            "permission": ['ANDROID_PERMISSION'],
            "pkg_name": ['ANDROID_PKG_NAME'],
            "provides_component": ['ANDROID_PROVIDES_COMPONENT'],
            "sdk": {
                "min": ['ANDROID_MINSDK'],
                "target": ['ANDROID_TARGET_SDK'],
            },
            "used_library": ['ANDROID_USE_LIBRARY'],
        },
        "jar": {
            "main_class": ['JAR_MAIN_CLASS'],
            "main_package": ['JAR_MAIN_PACKAGE']
        },
        "img": {
            "exif_tool": {
                "creator_tool": ['EXIFTOOL_XMP_CREATOR_TOOL'],
                "derived_document_id": ['EXIFTOOL_XMP_DERIVED_DOCUMENT_ID'],
                "document_id": ['EXIFTOOL_XMP_DOCUMENT_ID'],
                "instance_id": ['EXIFTOOL_XMP_INSTANCE_ID'],
                "toolkit": ['EXIFTOOL_XMP_TOOLKIT'],
            },
            "mega_pixels": ['IMAGE_MEGAPIXELS'],
            "mode": ['IMAGE_MODE'],
            "size": ['IMAGE_SIZE'],
            "sorted_metadata_hash": ['SORTED_METADATA_HASH'],
        },
        "ole": {
            "clsid": ['OLE_CLSID'],
            "dde_link": ['OLE_DDE_LINK'],
            "fib_timestamp": ['OLE_FIB_TIMESTAMP'],
            "macro": {
                "sha256": ['OLE_MACRO_SHA256'],
                "suspicious_string": ['OLE_MACRO_SUSPICIOUS_STRINGS'],
            },
            "summary": {
                "author": ['OLE_SUMMARY_AUTHOR'],
                "codepage": ['OLE_SUMMARY_CODEPAGE'],
                "comment": ['OLE_SUMMARY_COMMENTS'],
                "company": ['OLE_SUMMARY_COMPANY'],
                "create_time": ['OLE_SUMMARY_CREATETIME'],
                "last_printed": ['OLE_SUMMARY_LASTPRINTED'],
                "last_saved_by": ['OLE_SUMMARY_LASTSAVEDBY'],
                "last_saved_time": ['OLE_SUMMARY_LASTSAVEDTIME'],
                "manager": ['OLE_SUMMARY_MANAGER'],
                "subject": ['OLE_SUMMARY_SUBJECT'],
                "title": ['OLE_SUMMARY_TITLE'],
            },
        },
        "pe": {
            "api_vector": ['PE_APIVECTOR'],
            "exports": {
                "function_name": ['PE_EXPORT_FCT_NAME'],
                "module_name": ['PE_EXPORT_MODULE_NAME'],
            },
            "imports": {
                "fuzzy": ['PE_IMPORT_FUZZY'],
                "md5": ['PE_IMPORT_MD5'],
                "sorted_fuzzy": ['PE_IMPORT_SORTED_FUZZY'],
                "sorted_sha1": ['PE_IMPORT_SORTED_SHA1'],
                "suspicious": ['SUSPICIOUS_IMPORTS'],
            },
            "linker": {
                "timestamp": ['PE_LINK_TIME_STAMP'],
            },
            "oep": {
                "bytes": ['PE_OEP_BYTES'],
                "hexdump": ['PE_OEP_HEXDUMP'],
            },
            "pdb_filename": ['PE_PDB_FILENAME', 'FILE_PDB_STRING'],
            "resources": {
                "language": ['PE_RESOURCE_LANGUAGE'],
                "name": ['PE_RESOURCE_NAME'],
            },
            "sections": {
                "hash": ['PE_SECTION_HASH'],
                "name": ['PE_SECTION_NAME', 'PE_UNEXPECTED_SECTION_NAME'],
            },
            "version": {
                "description": ['PE_VERSION_INFO_FILE_DESCRIPTION'],
                "filename": ['PE_VERSION_INFO_ORIGINAL_FILENAME'],
            },
        },
        "pdf": {
            "date": {
                "modified": ['PDF_DATE_MOD'],
                "pdfx": ['PDF_DATE_PDFX'],
                "source_modified": ['PDF_DATE_SOURCEMODIFIED'],
            },
            "javascript": {
                "sha1": ['PDF_JAVASCRIPT_SHA1']
            },
            "stats": {
                "sha1": ['PDF_STATS_SHA1']
            },
        },
        "plist": {
            "installer_url": ['PLIST_APINSTALLERURL'],
            "build": {
                "machine_os": ['PLIST_BUILDMACHINEOSBUILD']
            },
            "cf_bundle": {
                "development_region": ['PLIST_CFBUNDLEDEVELOPMENTREGION'],
                "display_name": ['PLIST_CFBUNDLEDISPLAYNAME'],
                "executable": ['PLIST_CFBUNDLEEXECUTABLE'],
                'identifier': ['PLIST_CFBUNDLEIDENTIFIER'],
                'name': ['PLIST_CFBUNDLENAME'],
                "pkg_type": ['PLIST_CFBUNDLEPACKAGETYPE'],
                "signature": ['PLIST_CFBUNDLESIGNATURE'],
                "url_scheme": ['PLIST_CFBUNDLEURLSCHEMES'],
                "version": {
                    'long': ['PLIST_CFBUNDLEVERSION'],
                    'short': ['PLIST_CFBUNDLESHORTVERSIONSTRING'],
                }
            },
            "dt": {
                "compiler": ['PLIST_DTCOMPILER'],
                "platform": {
                    "build": ['PLIST_DTPLATFORMBUILD'],
                    "name": ['PLIST_DTPLATFORMNAME'],
                    "version": ['PLIST_DTPLATFORMVERSION'],
                }
            },
            "ls": {
                "background_only": ['PLIST_LSBACKGROUNDONLY'],
                "min_system_version": ['PLIST_LSMINIMUMSYSTEMVERSION'],
            },
            "min_os_version": ['PLIST_MINIMUMOSVERSION'],
            "ns": {
                "apple_script_enabled": ['PLIST_NSAPPLESCRIPTENABLED'],
                "principal_class": ['PLIST_NSPRINCIPALCLASS'],
            },
            "requests_open_access": ['PLIST_REQUESTSOPENACCESS'],
            "ui": {
                "background_modes": ['PLIST_UIBACKGROUNDMODES'],
                "requires_persistent_wifi": ['PLIST_UIREQUIRESPERSISTENTWIFI']
            },
            "wk": {
                "app_bundle_identifier": ['PLIST_WKAPPBUNDLEIDENITIFER'],
            }
        },
        "powershell": {
            "cmdlet": ['POWERSHELL_CMDLET']
        },
        "swf": {
            "header": {
                "frame": {
                    "size": ['SWF_HEADER_FRAME_SIZE'],
                    "rate": ['SWF_HEADER_FRAME_RATE'],
                    "count": ['SWF_HEADER_FRAME_COUNT'],
                },
                "version": ['SWF_HEADER_VERSION'],
            },
            "tags_ssdeep": ['SWF_TAGS_SSDEEP'],
        },
    },
    "network": {
        "attack": ['NET_ATTACK'],
        "domain": ['NET_DOMAIN_NAME'],
        "email": {
            "address": ['NET_EMAIL'],
            "date": ['NET_EMAIL_DATE'],
            "subject": ['NET_EMAIL_SUBJECT'],
            "msg_id": ['NET_EMAIL_MSG_ID'],
        },
        "ip": ['NET_IP'],
        "port": ['NET_PORT'],
        "protocol": ['NET_PROTOCOL', 'NET_PROTOCOL_SUSPICIOUS'],
        "signature": {
            "signature_id": ['SURICATA_SIGNATURE_ID'],
            "message": ['SURICATA_SIGNATURE_MESSAGE']
        },
        "tls": {
            "ja3_hash": ['TLS_JA3_HASH'],
            "ja3_string": ['TLS_JA3_STRING'],
        },
        "uri": ['NET_FULL_URI'],
        "uri_path": ['NET_NO_DOMAIN_URI'],
    },
    "source": ['SOURCE'],
    "technique": {
        "comms_routine": ['TECHNIQUE_COMMS_ROUTINE'],
        "config": ['TECHNIQUE_CONFIG'],
        "crypto": ['TECHNIQUE_CRYPTO'],
        "keylogger": ['TECHNIQUE_KEYLOGGER'],
        "macro": ['TECHNIQUE_MACROS'],
        "masking_algo": ['MASKING_ALGO'],
        "obfuscation": ['TECHNIQUE_OBFUSCATION', 'FILE_OBFUSCATION'],
        "packer": ['TECHNIQUE_PACKER'],
        "persistence": ['TECHNIQUE_PERSISTENCE'],
        "shellcode": ['SHELLCODE', 'TECHNIQUE_SHELLCODE'],
        "string": ['IMPLANT_STRINGS'],
    },
}

UNUSED = [
    'BASE64_ALPHABET',
    'DISPLAY_SEARCH_STRING',
    'DYNAMIC_MALICIOUSNESS',
    'DYNAMIC_MALWARE_PATTERN',
    'FILE_ATTRIBUTION',
    'FILE_EXTENSION',
    'FILE_MIMETYPE',
    'HEURISTIC',
    'REQUEST_SCORE',
    'REQUEST_USERNAME',
    'SERVICE_DESCRIPTION',
    'SERVICE_DISPLAY_NAME',
    'SERVICE_NAME',
]


def reverse_map(data: Dict) -> Dict:
    output = {}
    for k, v in data.items():
        for x in v:
            output[x] = k

    return output


v3_lookup_map = reverse_map(flatten(tag_map))
v3_lookup_map.update({k: None for k in UNUSED})



import os
import tempfile

from cart import pack_stream, unpack_stream, is_cart
from assemblyline.common import identify


# noinspection PyBroadException
def decode_file(original_path, fileinfo):
    extracted_path = None
    hdr = {}
    with open(original_path, 'rb') as original_file:
        if is_cart(original_file.read(256)):
            original_file.seek(0)

            extracted_fd, extracted_path = tempfile.mkstemp()
            extracted_file = os.fdopen(extracted_fd, 'wb')

            cart_extracted = False
            try:
                hdr, _ = unpack_stream(original_file, extracted_file)
                cart_extracted = True

            except Exception:
                extracted_path = None
                hdr = {}
                fileinfo['type'] = 'corrupted/cart'

            finally:
                extracted_file.close()

            if cart_extracted:
                fileinfo = identify.fileinfo(extracted_path)

    return extracted_path, fileinfo, hdr


# noinspection PyUnusedLocal
def encode_file(input_path, name, metadata=None):
    if metadata is None:
        metadata = {}

    _, output_path = tempfile.mkstemp()

    with open(output_path, 'wb') as oh:
        with open(input_path, 'rb') as ih:
            data = ih.read(64)
            if not is_cart(data):
                ih.seek(0)
                metadata.update({'name': name})
                pack_stream(ih, oh, metadata)
                return output_path, f"{name}.cart"
            else:
                return input_path, name



import smtplib

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from notifications_python_client.notifications import NotificationsAPIClient
from textwrap import dedent

from assemblyline.common import forge

config = forge.get_config()


def send_email(title: str, message: str, to: str):
    # set up the SMTP server
    s = smtplib.SMTP(host=config.auth.internal.signup.smtp.host, port=config.auth.internal.signup.smtp.port)
    if config.auth.internal.signup.smtp.tls:
        s.starttls()
    s.login(config.auth.internal.signup.smtp.user, config.auth.internal.signup.smtp.password)

    # For each contact, send the email:
    msg = MIMEMultipart()  # create a message

    # setup the parameters of the message
    msg['From'] = config.auth.internal.signup.smtp.from_adr
    msg['To'] = to
    msg['Subject'] = title

    # add in the message body
    msg.attach(MIMEText(message, 'plain'))

    text = msg.as_string()
    s.sendmail(config.auth.internal.signup.smtp.from_adr, to, text)

    # Terminate the SMTP session and close the connection
    s.quit()


def send_reset_email(to: str, reset_id: str):
    if config.auth.internal.signup.notify.base_url is not None:
        nc = NotificationsAPIClient(config.auth.internal.signup.notify.api_key,
                                    base_url=config.auth.internal.signup.notify.base_url)
        nc.send_email_notification(to, config.auth.internal.signup.notify.password_reset_template,
                                   personalisation={"fqdn": config.ui.fqdn, "reset_id": reset_id})
    else:
        message = dedent(f"""   
        We have received a request to have your password reset for {config.ui.fqdn}.
        
        To reset your password, please visit the link below: 
        
        https://{config.ui.fqdn}/?reset_id={reset_id}
        
        If you did not make this request, you can safely ignore this email and your password will remain the same.
        """)

        title = f"Assemblyline password reset request for {config.ui.fqdn}"

        send_email(title, message, to)


def send_signup_email(to: str, registration_key: str):
    if config.auth.internal.signup.notify.base_url is not None:
        nc = NotificationsAPIClient(config.auth.internal.signup.notify.api_key,
                                    base_url=config.auth.internal.signup.notify.base_url)
        nc.send_email_notification(to, config.auth.internal.signup.notify.registration_template,
                                   personalisation={"fqdn": config.ui.fqdn, "registration_key": registration_key})
    else:
        message = dedent(f"""
        We have received your request to register for {config.ui.fqdn}.
        
        To confirm your account registration, please visit the link below: 
        
        https://{config.ui.fqdn}/?registration_key={registration_key}
        
        If you did not make this request, you can safely ignore this email.
        """)

        title = f"Confirm your Assemblyline account registration for {config.ui.fqdn}"

        send_email(title, message, to)


def send_authorize_email(to: str, user: str, email: str):
    if config.auth.internal.signup.notify.base_url is not None:
        nc = NotificationsAPIClient(config.auth.internal.signup.notify.api_key,
                                    base_url=config.auth.internal.signup.notify.base_url)
        nc.send_email_notification(to, config.auth.internal.signup.notify.authorization_template,
                                   personalisation={"fqdn": config.ui.fqdn, "user": user, "email": email})
    else:
        message = dedent(f"""
        The following user has created an account and is waiting that his account gets activated.
        
        User: {user}
        Email: {email}
        
        You can browse to the link below to activate the account:
        https://{config.ui.fqdn}/admin/users?query=is_active%3Afalse
        """)

        title = f"A new {config.ui.fqdn} user is waiting for your authorization"

        send_email(title, message, to)


def send_activated_email(to: str, user: str, email: str, admin: str):
    if config.auth.internal.signup.notify.base_url is not None:
        nc = NotificationsAPIClient(config.auth.internal.signup.notify.api_key,
                                    base_url=config.auth.internal.signup.notify.base_url)
        nc.send_email_notification(to, config.auth.internal.signup.notify.activated_template,
                                   personalisation={"fqdn": config.ui.fqdn, "user": user,
                                                    "admin": admin, "email": email})
    else:
        message = dedent(f"""
        The following account was activated on {config.ui.fqdn}.
        
        Username: {user}
        Email: {email}

        The account was activated by: {admin}
        
        Login at https://{config.ui.fqdn}/
        """)

        title = f"Account for {user} now active on {config.ui.fqdn}"

        send_email(title, message, to)



from assemblyline import odm
from assemblyline.common import forge
Classification = forge.get_classification()


@odm.model(index=True, store=False)
class SubmissionTree(odm.Model):
    classification = odm.Classification(default=Classification.UNRESTRICTED)  # Classification of the cache
    filtered = odm.Boolean(default=False)                                     # Has this cache entry been filtered
    expiry_ts = odm.Date()        # Expiry date
    tree = odm.Text(index=False)  # Tree cache



from assemblyline import odm


@odm.model(index=False, store=False)
class FileScore(odm.Model):
    psid = odm.Optional(odm.UUID())      # ID of the parent submission to the associated submission
    expiry_ts = odm.Date(index=True)     # Expiry timestamp, used for garbage collection.
    score = odm.Integer()                # Maximum score for the associated submission
    errors = odm.Integer()               # Number of errors that occurred during the previous analysis
    sid = odm.UUID()                     # ID of the associated submission
    time = odm.Float()                   # Epoch time at which the FileScore entry was created



from assemblyline import odm
from assemblyline.common import forge
Classification = forge.get_classification()


PRIORITIES = {None, "LOW", "MEDIUM", "HIGH", "CRITICAL"}
STATUSES = {None, "MALICIOUS", "NON-MALICIOUS", "ASSESS", "TRIAGE"}


@odm.model(index=True, store=True)
class Workflow(odm.Model):
    classification = odm.Classification(
        copyto="__text__",
        default=Classification.UNRESTRICTED)                                 # Classification of the workflow
    creation_date = odm.Date(default="NOW")                                  # Creation date of the workflow
    creator = odm.Keyword()                                                  # UID of the creator
    edited_by = odm.Keyword()                                                # UID of the last edit user
    hit_count = odm.Integer(default=0)                                       # Number of time workflow hit
    labels = odm.List(odm.Keyword(), copyto="__text__", default=[])          # Labels applied by the workflow
    last_edit = odm.Date(default="NOW")                                      # Last edit date
    last_seen = odm.Optional(odm.Date())                                     # Last hit date
    name = odm.Keyword(copyto="__text__")                                    # Name of the workflow
    priority = odm.Optional(odm.Enum(copyto="__text__", values=PRIORITIES))  # Priority applied by the workflow
    query = odm.Keyword()                                                    # Query that the workflow runs
    status = odm.Optional(odm.Enum(copyto="__text__", values=STATUSES))      # Status applied by the workflow
    workflow_id = odm.Optional(odm.UUID())                                   # ID of the workflow



from assemblyline import odm
from assemblyline.common import forge
from assemblyline.common.caching import generate_conf_key
from assemblyline.odm.models.tagging import Tagging


BODY_FORMAT = {"TEXT", "MEMORY_DUMP", "GRAPH_DATA", "URL", "JSON", "KEY_VALUE", "PROCESS_TREE", "TABLE"}
constants = forge.get_constants()


@odm.model(index=True, store=False)
class Attack(odm.Model):
    attack_id = odm.Keyword(copyto="__text__")  # Attack matrix ID
    pattern = odm.Keyword(copyto="__text__")                        # Attack matrix Pattern Name
    categories = odm.List(odm.Keyword())                            # Attack matrix Categories


@odm.model(index=True, store=False)
class Signature(odm.Model):
    name = odm.Keyword(copyto="__text__")   # Name of the signature that triggered the heuristic
    frequency = odm.Integer(default=1)      # Number of times this signature triggered the heuristic


@odm.model(index=True, store=False)
class Heuristic(odm.Model):
    heur_id = odm.Keyword(copyto="__text__")                    # ID of th heuristic triggered
    name = odm.Keyword(copyto="__text__")                       # Name of the heuristic
    attack = odm.List(odm.Compound(Attack), default=[])         # List of Att&ck IDs related to this heuristic
    signature = odm.List(odm.Compound(Signature), default=[])   # List of signatures that triggered the heuristic
    score = odm.Integer()                                       # Computed Heuristic's score


@odm.model(index=True, store=False)
class Section(odm.Model):
    body = odm.Optional(odm.Text(copyto="__text__"))         # Text body of the result section
    classification = odm.Classification()                    # Classification of the section
    body_format = odm.Enum(values=BODY_FORMAT, index=False)  # Type of body in this section
    depth = odm.Integer(index=False)                         # Depth of the section
    heuristic = odm.Optional(odm.Compound(Heuristic))        # Heuristic used to score result section
    tags = odm.Compound(Tagging, default={})                 # List of tags associated to this section
    title_text = odm.Text(copyto="__text__")                 # Title of the section


@odm.model(index=True, store=True)
class ResultBody(odm.Model):
    score = odm.Integer(default=0)                          # Aggregate of the score for all heuristics
    sections = odm.List(odm.Compound(Section), default=[])  # List of sections


@odm.model(index=False, store=False)
class Milestone(odm.Model):
    service_started = odm.Date(default="NOW")    # Date the service started scanning
    service_completed = odm.Date(default="NOW")  # Date the service finished scanning


@odm.model(index=True, store=False)
class File(odm.Model):
    name = odm.Keyword(copyto="__text__")      # Name of the file
    sha256 = odm.SHA256(copyto="__text__")     # SHA256 hash of the file
    description = odm.Text(copyto="__text__")  # Description of the file
    classification = odm.Classification()      # Classification of the file


@odm.model(index=True, store=True)
class ResponseBody(odm.Model):
    milestones = odm.Compound(Milestone, default={})                          # Milestone block
    service_version = odm.Keyword(store=False)                                # Version of the service
    service_name = odm.Keyword(copyto="__text__")                             # Name of the service that scan the file
    service_tool_version = odm.Optional(odm.Keyword(copyto="__text__"))       # Tool version of the service
    supplementary = odm.List(odm.Compound(File), default=[])                  # List of supplementary files
    extracted = odm.List(odm.Compound(File), default=[])                      # List of extracted files
    service_context = odm.Optional(odm.Keyword(index=False, store=False))     # Context about the service
    service_debug_info = odm.Optional(odm.Keyword(index=False, store=False))  # Debug info about the service


@odm.model(index=True, store=True)
class Result(odm.Model):
    archive_ts = odm.Date(store=False)                         # Archiving timestamp
    classification = odm.Classification()                      # Aggregate classification for the result
    created = odm.Date(default="NOW")                          # Date at which the result object got created
    expiry_ts = odm.Optional(odm.Date(store=False))            # Expiry time stamp
    response: ResponseBody = odm.Compound(ResponseBody)        # The body of the response from the service
    result: ResultBody = odm.Compound(ResultBody, default={})  # The result body
    sha256 = odm.SHA256(store=False)                           # SHA256 of the file the result object relates to
    drop_file = odm.Boolean(default=False)                     # Do not pass to other stages after this run

    def build_key(self, service_tool_version=None, task=None):
        return self.help_build_key(
            self.sha256,
            self.response.service_name,
            self.response.service_version,
            self.is_empty(),
            service_tool_version=service_tool_version,
            task=task
        )

    @staticmethod
    def help_build_key(sha256, service_name, service_version, is_empty, service_tool_version=None, task=None):
        key_list = [
            sha256,
            service_name.replace('.', '_'),
            f"v{service_version.replace('.', '_')}",
            f"c{generate_conf_key(service_tool_version=service_tool_version, task=task)}",
        ]

        if is_empty:
            key_list.append("e")

        return '.'.join(key_list)

    def is_empty(self):
        if len(self.response.extracted) == 0 and \
                len(self.response.supplementary) == 0 and \
                len(self.result.sections) == 0 and \
                self.result.score == 0:
            return True
        return False




import json
import os
import random
import time
import threading

from multiprocessing import Process

from assemblyline.common import forge
from assemblyline.common.uid import get_random_id
from assemblyline.odm.models.error import ERROR_TYPES
from assemblyline.remote.datatypes.hash import Hash
from assemblyline.remote.datatypes.queues.named import NamedQueue


# noinspection PyBroadException
def backup_worker(worker_id, instance_id, working_dir):
    datastore = forge.get_datastore(archive_access=True)
    worker_queue = NamedQueue(f"r-worker-{instance_id}", ttl=1800)
    done_queue = NamedQueue(f"r-done-{instance_id}", ttl=1800)
    hash_queue = Hash(f"r-hash-{instance_id}")
    stopping = False
    with open(os.path.join(working_dir, "backup.part%s" % worker_id), "w+") as backup_file:
        while True:
            data = worker_queue.pop(timeout=1)
            if data is None:
                if stopping:
                    break
                continue

            if data.get('stop', False):
                if not stopping:
                    stopping = True
                else:
                    time.sleep(round(random.uniform(0.050, 0.250), 3))
                    worker_queue.push(data)
                continue

            missing = False
            success = True
            try:
                to_write = datastore.get_collection(data['bucket_name']).get(data['key'], as_obj=False)
                if to_write:
                    if data.get('follow_keys', False):
                        for bucket, bucket_key, getter in FOLLOW_KEYS.get(data['bucket_name'], []):
                            for key in getter(to_write.get(bucket_key, None)):
                                hash_key = "%s_%s" % (bucket, key)
                                if not hash_queue.exists(hash_key):
                                    hash_queue.add(hash_key, "True")
                                    worker_queue.push({"bucket_name": bucket, "key": key, "follow_keys": True})

                    backup_file.write(json.dumps((data['bucket_name'], data['key'], to_write)) + "\n")
                else:
                    missing = True
            except Exception:
                success = False

            done_queue.push({
                "success": success,
                "missing": missing,
                "bucket_name": data['bucket_name'],
                "key": data['key']
            })

    done_queue.push({"stopped": True})


# noinspection PyBroadException
def restore_worker(worker_id, instance_id, working_dir):
    datastore = forge.get_datastore(archive_access=True)
    done_queue = NamedQueue(f"r-done-{instance_id}", ttl=1800)

    with open(os.path.join(working_dir, "backup.part%s" % worker_id), "rb") as input_file:
        for line in input_file:
            bucket_name, key, data = json.loads(line)

            success = True
            try:
                collection = datastore.get_collection(bucket_name)
                collection.save(key, data)
            except Exception:
                success = False

            done_queue.push({
                "success": success,
                "missing": False,
                "bucket_name": bucket_name,
                "key": key})

    done_queue.push({"stopped": True})


class DistributedBackup(object):
    def __init__(self, working_dir, worker_count=50, spawn_workers=True, use_threading=False, logger=None):
        self.working_dir = working_dir
        self.datastore = forge.get_datastore(archive_access=True)
        self.logger = logger
        self.plist = []
        self.use_threading = use_threading
        self.instance_id = get_random_id()
        self.worker_queue = NamedQueue(f"r-worker-{self.instance_id}", ttl=1800)
        self.done_queue = NamedQueue(f"r-done-{self.instance_id}", ttl=1800)
        self.hash_queue = Hash(f"r-hash-{self.instance_id}")
        self.bucket_error = []
        self.VALID_BUCKETS = sorted(list(self.datastore.ds.get_models().keys()))
        self.worker_count = worker_count
        self.spawn_workers = spawn_workers
        self.total_count = 0
        self.error_map_count = {}
        self.missing_map_count = {}
        self.map_count = {}
        self.last_time = 0
        self.last_count = 0
        self.error_count = 0

    def cleanup(self):
        self.worker_queue.delete()
        self.done_queue.delete()
        self.hash_queue.delete()
        for p in self.plist:
            p.terminate()

    def done_thread(self, title):
        t0 = time.time()
        self.last_time = t0

        running_threads = self.worker_count

        while running_threads > 0:
            msg = self.done_queue.pop(timeout=1)

            if msg is None:
                continue

            if "stopped" in msg:
                running_threads -= 1
                continue

            bucket_name = msg.get('bucket_name', 'unknown')

            if msg.get('success', False):
                self.total_count += 1

                if msg.get("missing", False):
                    if bucket_name not in self.missing_map_count:
                        self.missing_map_count[bucket_name] = 0

                    self.missing_map_count[bucket_name] += 1
                else:
                    if bucket_name not in self.map_count:
                        self.map_count[bucket_name] = 0

                    self.map_count[bucket_name] += 1

                new_t = time.time()
                if (new_t - self.last_time) > 5:
                    if self.logger:
                        self.logger.info("%s (%s at %s keys/sec) ==> %s" %
                                         (self.total_count,
                                          new_t - self.last_time,
                                          int((self.total_count - self.last_count) / (new_t - self.last_time)),
                                          self.map_count))
                    self.last_count = self.total_count
                    self.last_time = new_t
            else:
                self.error_count += 1

                if bucket_name not in self.error_map_count:
                    self.error_map_count[bucket_name] = 0

                self.error_map_count[bucket_name] += 1

        # Cleanup
        self.cleanup()

        summary = ""
        summary += "\n########################\n"
        summary += "####### SUMMARY  #######\n"
        summary += "########################\n"
        summary += "%s items - %s errors - %s secs\n\n" % \
                   (self.total_count, self.error_count, time.time() - t0)

        for k, v in self.map_count.items():
            summary += "\t%15s: %s\n" % (k.upper(), v)

        if len(self.missing_map_count.keys()) > 0:
            summary += "\n\nMissing data:\n\n"
            for k, v in self.missing_map_count.items():
                summary += "\t%15s: %s\n" % (k.upper(), v)

        if len(self.error_map_count.keys()) > 0:
            summary += "\n\nErrors:\n\n"
            for k, v in self.error_map_count.items():
                summary += "\t%15s: %s\n" % (k.upper(), v)

        if len(self.bucket_error) > 0:
            summary += f"\nThese buckets failed to {title.lower()} completely: {self.bucket_error}\n"
        if self.logger:
            self.logger.info(summary)

    # noinspection PyBroadException,PyProtectedMember
    def backup(self, bucket_list, follow_keys=False, query=None):
        if query is None:
            query = 'id:*'

        for bucket in bucket_list:
            if bucket not in self.VALID_BUCKETS:
                if self.logger:
                    self.logger.warn("\n%s is not a valid bucket.\n\n"
                                     "The list of valid buckets is the following:\n\n\t%s\n" %
                                     (bucket.upper(), "\n\t".join(self.VALID_BUCKETS)))
                return

        targets = ', '.join(bucket_list)
        try:
            if self.logger:
                self.logger.info("\n-----------------------")
                self.logger.info("----- Data Backup -----")
                self.logger.info("-----------------------")
                self.logger.info(f"    Deep: {follow_keys}")
                self.logger.info(f"    Buckets: {targets}")
                self.logger.info(f"    Workers: {self.worker_count}")
                self.logger.info(f"    Target directory: {self.working_dir}")
                self.logger.info(f"    Filtering query: {query}")

            # Start the workers
            for x in range(self.worker_count):
                if self.use_threading:
                    t = threading.Thread(target=backup_worker, args=(x, self.instance_id, self.working_dir))
                    t.setDaemon(True)
                    t.start()
                else:
                    p = Process(target=backup_worker, args=(x, self.instance_id, self.working_dir))
                    p.start()
                    self.plist.append(p)

            # Start done thread
            dt = threading.Thread(target=self.done_thread, args=('Backup',), name="Done thread")
            dt.setDaemon(True)
            dt.start()

            # Process data buckets
            for bucket_name in bucket_list:
                try:
                    collection = self.datastore.get_collection(bucket_name)
                    for item in collection.stream_search(query, fl="id", item_buffer_size=500, as_obj=False):
                        self.worker_queue.push({"bucket_name": bucket_name, "key": item['id'],
                                                "follow_keys": follow_keys})

                except Exception as e:
                    self.cleanup()
                    if self.logger:
                        self.logger.execption(e)
                        self.logger.error("Error occurred while processing bucket %s." % bucket_name)
                    self.bucket_error.append(bucket_name)

            for _ in range(self.worker_count):
                self.worker_queue.push({"stop": True})

            dt.join()
        except Exception as e:
            if self.logger:
                self.logger.execption(e)

    def restore(self):
        try:
            if self.logger:
                self.logger.info("\n------------------------")
                self.logger.info("----- Data Restore -----")
                self.logger.info("------------------------")
                self.logger.info(f"    Workers: {self.worker_count}")
                self.logger.info(f"    Target directory: {self.working_dir}")

            for x in range(self.worker_count):
                if self.use_threading:
                    t = threading.Thread(target=restore_worker,
                                         args=(x, self.instance_id, self.working_dir))
                    t.setDaemon(True)
                    t.start()
                else:
                    p = Process(target=restore_worker, args=(x, self.instance_id, self.working_dir))
                    p.start()
                    self.plist.append(p)

            # Start done thread
            dt = threading.Thread(target=self.done_thread, args=('Restore',), name="Done thread")
            dt.setDaemon(True)
            dt.start()

            # Wait for workers to finish
            dt.join()
        except Exception as e:
            if self.logger:
                self.logger.execption(e)


def _string_getter(data):
    if data is not None:
        return [data]
    else:
        return []


def _result_getter(data):
    if data is not None:
        return [x for x in data if not x.endswith('.e')]
    else:
        return []


def _emptyresult_getter(data):
    if data is not None:
        return [x for x in data if x.endswith('.e')]
    else:
        return []


def _error_getter(data):
    if data is not None:
        return [x for x in data if x.rsplit('.e', 1)[1] not in ERROR_TYPES.values()]
    else:
        return []


def _sha256_getter(data):
    if data is not None:
        return [x[:64] for x in data]
    else:
        return []


def _file_getter(data):
    if data is not None:
        return [x['sha256'] for x in data]
    else:
        return []


def _result_file_getter(data):
    if data is not None:
        supp = data.get("supplementary", []) + data.get("extracted", [])
        return _file_getter(supp)
    else:
        return []


FOLLOW_KEYS = {
    "alert": [
        ('submission', 'sid', _string_getter),
    ],
    "submission": [
        ('result', 'results', _result_getter),
        ('error', 'errors', _error_getter),
        ('file', 'results', _sha256_getter),
        ('file', 'files', _file_getter),
        ('file', 'errors', _sha256_getter),
    ],
    "results": [
        ('file', 'response', _result_file_getter),
    ]
}
